import tempfile

import h5py
import numpy as np
import pytest

from nexgen.nxs_utils import Axis, TransformationType
from nexgen.tools.Metafile import DectrisMetafile, TristanMetafile
from nexgen.tools.MetaReader import (
    define_vds_data_type,
    overwrite_beam,
    update_axes_from_meta,
)

axes_list = [
    Axis("omega", ".", TransformationType.ROTATION, (0, 0, -1)),
    Axis("sam_z", "omega", TransformationType.TRANSLATION, (0, 0, 1)),
    Axis("sam_y", "sam_z", TransformationType.TRANSLATION, (0, 1, 0)),
    Axis("sam_x", "sam_y", TransformationType.TRANSLATION, (1, 0, 0)),
    Axis("phi", "sam_x", TransformationType.ROTATION, (0, 0, 1)),
]

test_detector_size = (512, 1028)  # slow, fast
test_beam = {"wavelength": 0.0}

test_detector = {"axes": ["two_theta", "det_z"]}

dummy_config = """{
    "nimages": 10,
    "ntrigger": 1,
    "omega_increment": 0.0,
    "omega_start": 90.0,
    "phi_increment": 0.1,
    "phi_start": 0.0,
    }"""


def test_Tristan_meta_file():
    test_hdf_file = tempfile.NamedTemporaryFile(suffix=".h5", delete=True)
    with h5py.File(test_hdf_file, "w") as test_meta_file:
        test_meta_file["ts_qty_module00"] = np.zeros(2)
        test_meta_file["ts_qty_module01"] = np.zeros(2)
        assert TristanMetafile(test_meta_file).find_number_of_modules() == 2
        assert TristanMetafile(test_meta_file).find_software_version() is None

    assert TristanMetafile.isTristan(test_hdf_file.name)


@pytest.fixture
def dummy_eiger_meta_file():
    test_hdf_file = tempfile.TemporaryFile()
    test_meta_file = h5py.File(test_hdf_file, "w")
    test_meta_file["config"] = dummy_config
    test_meta_file["_dectris/nimages"] = np.array([10])
    test_meta_file["_dectris/ntrigger"] = np.array([1])
    test_meta_file["_dectris/wavelength"] = np.array([0.6])
    test_meta_file["_dectris/x_pixels_in_detector"] = np.array([test_detector_size[1]])
    test_meta_file["_dectris/y_pixels_in_detector"] = np.array([test_detector_size[0]])
    test_meta_file["_dectris/phi_start"] = np.array([0.0])
    test_meta_file["_dectris/phi_increment"] = np.array([0.1])
    test_meta_file["_dectris/omega_start"] = np.array([90.0])
    test_meta_file["_dectris/omega_increment"] = np.array([0.0])
    test_meta_file["_dectris/two_theta_start"] = np.array([0.0])
    test_meta_file["_dectris/two_theta_increment"] = np.array([0.0])
    test_meta_file["_dectris/detector_distance"] = np.array([0.19])
    test_meta_file["_dectris/bit_depth_image"] = np.array([32])
    yield test_meta_file


def test_Eiger_meta_file(dummy_eiger_meta_file):
    meta = DectrisMetafile(dummy_eiger_meta_file)
    assert meta.hasDectrisGroup
    assert meta.hasMask is False and meta.hasFlatfield is False
    assert meta.read_dectris_config() == {
        "nimages": 10,
        "ntrigger": 1,
        "wavelength": 0.6,
        "x_pixels_in_detector": 1028,
        "y_pixels_in_detector": 512,
        "detector_distance": 0.19,
        "omega_increment": 0.0,
        "omega_start": 90.0,
        "phi_increment": 0.1,
        "phi_start": 0.0,
        "two_theta_increment": 0.0,
        "two_theta_start": 0.0,
        "bit_depth_image": 32,
    }
    assert meta.get_number_of_images() == 10
    assert meta.get_number_of_triggers() == 1
    assert meta.get_full_number_of_images() == 10
    assert meta.get_detector_size() == test_detector_size
    assert meta.hasConfig
    assert meta.read_config_dset() == {
        "nimages": 10,
        "ntrigger": 1,
        "omega_increment": 0.0,
        "omega_start": 90.0,
        "phi_increment": 0.1,
        "phi_start": 0.0,
    }


def test_overwrite_beam(dummy_eiger_meta_file):
    overwrite_beam(dummy_eiger_meta_file, "Eiger", test_beam)
    assert test_beam["wavelength"] == 0.6


def test_define_vds_shape(dummy_eiger_meta_file):
    meta = DectrisMetafile(dummy_eiger_meta_file)
    vds_shape = define_vds_data_type(meta)
    assert vds_shape is not None
    assert vds_shape == np.uint32


def test_update_axes_from_meta(dummy_eiger_meta_file):
    meta = DectrisMetafile(dummy_eiger_meta_file)
    assert axes_list[0].start_pos == 0.0  # omega
    assert axes_list[-1].start_pos == 0.0  # phi
    update_axes_from_meta(meta, axes_list, osc_axis="phi")
    assert axes_list[0].start_pos == 90.0  # omega
    assert axes_list[-1].start_pos == 0.0  # phi
    assert axes_list[-1].num_steps == 10


def test_update_axes_from_meta_using_config(dummy_eiger_meta_file):
    meta = DectrisMetafile(dummy_eiger_meta_file)
    # Reset axes
    axes_list[0].start_pos = 0.0
    axes_list[-1].start_pos = 0.0
    update_axes_from_meta(meta, axes_list, osc_axis="phi", use_config=True)
    assert axes_list[0].start_pos == 90.0  # omega
    assert axes_list[-1].start_pos == 0.0  # phi
    assert axes_list[-1].num_steps == 10



"""
Tools to write Virtual DataSets
"""
from __future__ import annotations

import logging
import operator
from dataclasses import dataclass
from functools import reduce
from pathlib import Path
from typing import List, Tuple

import h5py
import numpy as np
from numpy.typing import DTypeLike

from ..utils import MAX_FRAMES_PER_DATASET
from .Constants import jungfrau_fill_value, jungfrau_gap_size, jungfrau_mod_size

vds_logger = logging.getLogger("nexgen.VDSWriter")


@dataclass
class Dataset:
    name: str

    # The full shape of the source, regardless of start index
    source_shape: Tuple[int]

    # The start index that we should start copying from
    start_index: int = 0

    # The point where we should stop copying. Defaults to maximum for an Eiger
    stop_index: int = MAX_FRAMES_PER_DATASET

    # The shape of the destination, including the start_index
    dest_shape: Tuple[int] = None

    def __post_init__(self):
        self.dest_shape = (
            self.source_shape[0] - self.start_index,
            *self.source_shape[1:],
        )
        if (
            self.stop_index < MAX_FRAMES_PER_DATASET
            or self.stop_index < self.source_shape[0]
        ):
            self.dest_shape = (
                self.dest_shape[0] - (self.source_shape[0] - self.stop_index),
                *self.source_shape[1:],
            )

    def __add__(self, x):
        """Returns a dataset that has the same start index and shape as if the two were appended to each other."""
        return Dataset(
            "",
            source_shape=(
                self.source_shape[0] + x.source_shape[0],
                *self.source_shape[1:],
            ),
            start_index=self.start_index + x.start_index,
            stop_index=self.stop_index + x.stop_index,
        )


def find_datasets_in_file(nxdata: h5py.Group) -> List:
    """
    Look for the source datasets in the NeXus file. Assumes that the source datasets are always h5py.ExternalLink.

    Args:
        nxdata (h5py.Group): Group where the data should be linked.

    Raises:
        KeyError: If no ExternalLinks to data are found in the group.

    Returns:
        dsets (List): The source datasets.
    """
    # FIXME for now this assumes that the source datasets are always links
    dsets = []
    for k in nxdata.keys():
        if isinstance(nxdata.get(k, getlink=True), h5py.ExternalLink):
            dsets.append(k)
    if not dsets:
        raise KeyError(
            f"No External Link datasets found in NeXus file under {nxdata.name}"
        )
    return dsets


def split_datasets(
    dsets,
    data_shape: Tuple[int, int, int],
    start_idx: int = 0,
    vds_shape: Tuple[int, int, int] = None,
) -> List[Dataset]:
    """
    Splits the full data shape and start index up into values per dataset,
    given that each dataset has a maximum size.

    Args:
        dsets (Dataset): The input datasets.
        data_shape (Tuple[int, int, int]): Shape of the data, usually defined as (num_frames, *image_size).
        start_idx (int, optional): The start point for the source data. Defaults to 0.
        vds_shape(Tuple, optional): Desired shape of the VDS, usually defined as (num_frames, *image_size). \
            The number of frames must be smaller or equal to the one in full_data_shape. Defaults to None.

    Raises:
        ValueError: If the passed start index value is higher than the dataset lenght.
        ValueError: It the passed start index value is negative.

    Returns:
        List[Dataset]: A list of datasets.
    """
    if start_idx > data_shape[0]:
        raise ValueError(
            f"Start index {start_idx} must be less than full dataset length {data_shape[0]}"
        )
    if start_idx < 0:
        raise ValueError("Start index must be positive")

    if not isinstance(data_shape[0], int):
        vds_logger.warning("Datashape not passed as int, will attempt to cast")

    if not isinstance(start_idx, int):
        vds_logger.warning("VDS start index not passed as int, will attempt to cast")

    if vds_shape and not isinstance(vds_shape[0], int):
        vds_logger.warning("VDS start index not passed as int, will attempt to cast")

    if vds_shape is None:
        vds_logger.info(
            "VDS shape not chosen, it will be calculated from the full data shape and the chosen start index."
        )
        vds_shape = (data_shape[0] - start_idx, *data_shape[1:])

    full_frames = int(data_shape[0])
    end_cut_frames = int(full_frames - vds_shape[0]) - int(start_idx)

    result = []
    for dset_name in dsets:
        dset = Dataset(
            name=dset_name,
            source_shape=(min(MAX_FRAMES_PER_DATASET, full_frames), *data_shape[1:]),
            start_index=min(MAX_FRAMES_PER_DATASET, max(int(start_idx), 0)),
            stop_index=min(MAX_FRAMES_PER_DATASET, (full_frames - end_cut_frames)),
        )
        # if start index == 1000 then that source dataset is not used and we should
        # not pass it on to use as a source for the VDS
        # Same goes if all datasets from last files are not used
        if dset.start_index != MAX_FRAMES_PER_DATASET:
            if dset.stop_index > 0 and dset.stop_index <= MAX_FRAMES_PER_DATASET:
                result.append(dset)
        start_idx -= MAX_FRAMES_PER_DATASET
        full_frames -= MAX_FRAMES_PER_DATASET

    return result


def create_virtual_layout(datasets: List[Dataset], data_type: DTypeLike):
    """
    Create a virtual layout and populate it based on the provided data.

    Args:
        datasets (List[Dataset]): A list of datasets that are to be merged.
        data_type (DTypeLike): The type of the input data.

    Returns:
        layout (h5py.VirtualLayout): Virtual layout.
    """
    full_dataset: Dataset = reduce(operator.add, datasets)
    layout = h5py.VirtualLayout(shape=full_dataset.dest_shape, dtype=data_type)

    dest_start = 0
    for dataset in datasets:
        if dataset.stop_index == dataset.source_shape[0]:
            dest_end = dest_start + dataset.source_shape[0] - dataset.start_index
        else:
            dest_end = dest_start + dataset.dest_shape[0]

        vsource = h5py.VirtualSource(
            ".", "/entry/data/" + dataset.name, shape=dataset.source_shape
        )

        layout[dest_start:dest_end, :, :] = vsource[
            dataset.start_index : dataset.stop_index, :, :
        ]
        dest_start = dest_end

    return layout


def image_vds_writer(
    nxsfile: h5py.File,
    full_data_shape: Tuple | List,
    start_index: int = 0,
    vds_shape: Tuple | List | None = None,
    data_type: DTypeLike = np.uint16,
    entry_key: str = "data",
):
    """
    Virtual DataSet writer function for image data.

    Args:
        nxsfile (h5py.File): Handle to NeXus file being written.
        full_data_shape (Tuple | List): Shape of the full dataset, usually defined as (num_frames, *image_size).
        start_index(int): The start point for the source data. Defaults to 0.
        vds_shape(Tuple, optional): Desired shape of the VDS, usually defined as (num_frames, *image_size). \
            The number of frames must be smaller or equal to the one in full_data_shape. Defaults to None.
        data_type (DTypeLike, optional): The type of the input data. Defaults to np.uint16.
        entry_key (str, optional): Entry key for the Virtual DataSet name. Defaults to data.
    """
    vds_logger.info("Start creating VDS ...")
    # Where the vds will go
    nxdata = nxsfile["/entry/data"]
    dset_names = find_datasets_in_file(nxdata)

    vds_shape = (
        tuple(vds_shape)
        if vds_shape is not None
        else (full_data_shape[0] - start_index, *full_data_shape[1:])
    )

    # Hack for datasets with no maximum number of frames (eg. Singla)
    if len(dset_names) == 1 and full_data_shape[0] > MAX_FRAMES_PER_DATASET:
        # nxdata[dset_names[0]].shape[0] > MAX_FRAMES_PER_DATASET
        # .maxshape[0] is None
        datasets = [
            Dataset(
                name=dset_names[0],
                source_shape=full_data_shape,
                start_index=start_index,
                stop_index=full_data_shape[0],
            )
        ]
    else:
        datasets = split_datasets(dset_names, full_data_shape, start_index, vds_shape)

    layout = create_virtual_layout(datasets, data_type)

    # Writea Virtual Dataset in NeXus file
    nxdata.create_virtual_dataset(entry_key, layout, fillvalue=-1)
    vds_logger.info("VDS written to NeXus file.")


def jungfrau_vds_writer(
    nxsfile: h5py.File,
    vds_shape: Tuple | List,
    data_type: DTypeLike = np.uint16,
    source_dsets: List[str] | None = None,
):
    """Write VDS for Jungfrau 1M use case, with a tiled layout."""
    external_dsets = True
    entry_key = "data"
    frames = vds_shape[0]

    nxdata = nxsfile["/entry/data"]
    if not source_dsets:
        source_dsets = find_datasets_in_file(nxdata)
        external_dsets = False

    sources = []
    for dset in source_dsets:
        source_path = dset if external_dsets is True else "."
        source_name = entry_key if external_dsets is True else f"/entry/data/{dset}"
        source = h5py.VirtualSource(
            source_path, source_name, shape=(frames, *jungfrau_mod_size)
        )
        sources.append(source)

    layout = h5py.VirtualLayout(shape=vds_shape, dtype=data_type)
    # The first one is the upper one
    s0 = jungfrau_mod_size[0] + jungfrau_gap_size[0]
    layout[:, : jungfrau_mod_size[0], :] = sources[1][:, :, :]
    layout[:, s0:, :] = sources[0][:, :, :]

    nxdata.create_virtual_dataset(entry_key, layout, fillvalue=jungfrau_fill_value)


def vds_file_writer(
    nxsfile: h5py.File,
    datafiles: List[Path],
    data_shape: Tuple | List,
    data_type: DTypeLike = np.uint16,
    entry_key: str = "data",
):
    """
    Write a Virtual DataSet _vds.h5 file for image data.

    Args:
        nxsfile (h5py.File): NeXus file being written.
        datafiles (List[Path]): List of paths to source files.
        data_shape (Tuple | List): Shape of the dataset, usually defined as (num_frames, *image_size).
        data_type (DTypeLike, optional): Dtype. Defaults to np.uint16.
        entry_key (str): Entry key for the Virtual DataSet name. Defaults to data.
    """
    vds_logger.info("Start creating VDS ...")
    # Where the vds will go
    nxdata = nxsfile["/entry/data"]
    # entry_key = "data"

    # For every source dataset define its shape and number of frames
    # Once again, it is assumed that the maximum number of frames per dataset is 1000
    frames = (data_shape[0] // 1000) * [1000] + [data_shape[0] % 1000]
    sshape = [(f, *data_shape[1:]) for f in frames]

    # Create virtual layout
    layout = h5py.VirtualLayout(shape=data_shape, dtype=data_type)
    start = 0
    for n, filename in enumerate(datafiles):
        end = start + frames[n]
        vsource = h5py.VirtualSource(
            filename.name, entry_key, shape=sshape[n]
        )  # Source definition
        layout[start:end:1, :, :] = vsource
        start = end

    # Create a _vds.h5 file and add link to nexus file
    s = Path(nxsfile.filename).expanduser().resolve()
    vds_filename = s.parent / f"{s.stem}_vds.h5"
    del s
    with h5py.File(vds_filename, "w") as vds:
        vds.create_virtual_dataset("data", layout, fillvalue=-1)
    nxdata["data"] = h5py.ExternalLink(vds_filename.name, "data")
    vds_logger.info(f"{vds_filename} written and link added to NeXus file.")


def clean_unused_links(
    nxsfile: h5py.File,
    vds_shape: Tuple | List,
    start_index: int = 0,
):
    """
    Remove links to external data not used in VDS.

    Args:
        nxsfile (h5py.File): Handle to NeXus file being written.
        vds_shape (Tuple | List): Actual shape of the VDS dataset, usually defined as (num_frames, *image_size).
        start_index(int): The start point for the source data. Defaults to 0.
    """
    vds_logger.info("Cleaning links unused in VDS ...")
    # Location of the VDS
    nxdata = nxsfile["/entry/data"]
    dataset_names = find_datasets_in_file(nxdata)
    if len(dataset_names) == 1:
        vds_logger.info("Only one linked file, no need to remove it.")
        return
    datasets = [nxdata[name] for name in dataset_names]
    dataset_lengths = [d.shape[0] for d in datasets]
    if sum(dataset_lengths) == vds_shape[0]:
        vds_logger.info("All links are used in VDS, no need to remove any.")
        return
    for i, _ in enumerate(datasets):
        # unlink datasets before the start of VDS
        if sum(dataset_lengths[0 : i + 1]) < start_index:
            vds_logger.info(f"Removing {dataset_names[i]} link.")
            del nxdata[dataset_names[i]]
        # unlink datasets after the end of VDS
        if sum(dataset_lengths[0:i]) > start_index + vds_shape[0]:
            vds_logger.info(f"Removing {dataset_names[i]} link.")
            del nxdata[dataset_names[i]]
    vds_logger.info("Links unused in VDS removed from NeXus file.")



"""
Define phil scopes that describe beamline, goniometer, detector and module.
"""

import freephil

# TODO FIXME .multiple for fast and slow axis doesn't seem to work correctly
module_scope = freephil.parse(
    """
    detector_module {
      num_modules = 1
        .type = int
        .help = "Number of modules - defaults to 1."
      module_offset = 0 *1 2
        .type = choice
        .help = "Decide whether to calculate offset of the module in regard to detector origin and creates corresponding field.
                Options:
                0 -> do not calculate module_offset
                1 -> un-normalized displacement, set module_offset to 1
                2 -> normalized displacement and set module_offset to magnitude of displacement"
      fast_axis = None
        .type = floats(size = 3)
        .help = "Fast axis at datum position"
      slow_axis = None
        .type = floats(size = 3)
        .help = "Slow axis at datum position"
      offsets = None
        .type = floats
        .help = "Axis offsets - one after the other - fast then slow"
      module_size = None
        .type = ints
        .help = "In case of multiple modules, pass the size of each single module"
    }
    """
)

detector_scope = freephil.parse(
    """
    detector {
      mode = *images events
        .type = choice
        .help = "Detector acquisition mode. Defaults to images. Only relevant for Tristan collections."
      description = None
        .type = str
        .help = "Detector class to record"
      detector_type = Pixel
        .type = str
        .help = "Detector type to record"
      sensor_material = *Si CdTe
        .type = choice
        .help = "Sensor material (e.g. silicon)"
      sensor_thickness = 0.0mm
        .type = str
        .help = "Sensor thickness, if unit is not specified defaults to mm"
      overload = None
        .type = int
        .help = "Pixels >= this value are invalid due to overloading"
      underload = -1
        .type = int
        .help = "Pixels <= this value are invalid"
      pixel_size = 0.0mm 0.0mm
        .type = strings
        .help = "Pixel size, if unit isn't passed defaults to mm (fast, slow)"
      beam_center = 0.0 0.0
        .type = floats(size = 2)
        .help = "Beam position on the detector (fast, slow)"
      flatfield = None
        .type = path
        .help = "If path is given, add flatfield correction data field"
      flatfield_applied = False
        .type = bool
        .help = "If the flatfield location is known, specify whether it's been applied of not."
      pixel_mask = None
        .type = path
        .help = "if path is given, add link to bad pixel mask"
      pixel_mask_applied = False
        .type = bool
        .help = "If the mask location is known, specify whether it's been applied of not."
      image_size = 0 0
        .type = ints(size = 2)
        .help = "Image size in pixels: (slow, fast)"
      exposure_time = None
        .type = str
        .help = "Nominal exposure time, eg 0.004s, if unit is not specified defaults to seconds"
      axes = None
        .type = strings
        .help = "Axis names for the detector axes. The detector sits on the last one."
      depends = None
        .type = strings
        .help = "Axis names for the axis dependencies"
      vectors = None
        .type = floats
        .help = "Axis vectors - one after the other"
      starts = None
        .type = floats
        .help = "Starting values for axes"
      ends = None
        .type = floats
        .help = "Ending values for axes, only different to start for scan axis"
      increments = None
        .type = floats
        .help = "Increment values for axes, non-zero only for scan axis"
      types = None
        .type = strings
        .help = "Axis types, rotation or translation"
      units = None
        .type = strings
        .help = "Axis units, from mm or deg"
      software_version = None
        .type = str
        .help = "Detector software version"
      bit_depth_readout = None
        .type = int
        .help = How many bits the electronics reads per pixel.
    }
    tristanSpec {
      detector_tick = 1562.5ps
        .type = str
        .help = "Tristan specific - detector tick, in ps"
      detector_frequency = 6.4e+08Hz
        .type = str
        .help = "Tristan specific - detector frequency, in Hz"
      timeslice_rollover = 18
        .type = int
        .help = "Tristan specific - timeslice rollover bits"
    }
    """
)

goniometer_scope = freephil.parse(
    """
    goniometer {
      axes = None
        .type = strings
        .help = "Axis names for the goniometer axes. Sample depends on the last one in the list"
      depends = None
        .type = strings
        .help = "Axis names for the axis dependencies"
      vectors = None
        .type = floats
        .help = "Axis vectors - one after the other"
      offsets = None
        .type = floats
        .help = "Axis offsets - one after the other"
      offset_units = None
        .type = strings
        .help = "Units of the axis offsets (usually mm)"
      starts = None
        .type = floats
        .help = "Starting values for axes"
      ends = None
        .type = floats
        .help = "Ending values for axes, only different to start for scan axis"
      increments = None
        .type = floats
        .help = "Increment values for axes, non-zero only for scan axis"
      types = None
        .type = strings
        .help = "Axis types, rotation or translation"
      units = None
        .type = strings
        .help = "Axis units, from mm or deg"
    }
    """
)

instrument_scope = freephil.parse(
    """
    source {
      name = Diamond Light Source
        .type = str
        .help = "Facility name"
      short_name = DLS
        .type = str
        .help = "Facility abbreviation"
      type = Synchrotron X-ray Source
        .type = str
        .help = "Facility type"
      beamline_name = None
        .type = str
        .help = "Beamline name"
      facility_id = None
        .type = str
        .help = "Identifier to use to define instrument name, when different from Diamond. Defaults to None"
      probe = None
        .type = str
        .help = "Type of radiation probe"
    }

    beam {
      wavelength = None
        .type = float
        .help = "Wavelength of incident beam, angstroms"
      flux = None
        .type = float
        .help = "Flux of incident beam, ph / s"
    }

    attenuator {
      transmission = None
        .type = float
        .help = "Attenuation of beam intensity"
    }
    pump_probe {
      pump_status = False
        .type = bool
        .help = "Pump probe experiment status."
      pump_exp = None
        .type = float
        .help = "Pump exposure time"
      pump_delay = None
        .type = float
        .help = "Pump delay"
    }
    """
)

timestamp_scope = freephil.parse(
    """
    start_time = None
      .type = str
      .help = "Experiment start time, pass either a timestamp or a string, eg '2021-09-20T10:20:30' or 'Tue Sep 28 2021 10:58:01'."
    end_time = None
      .type = str
      .help = "Experiment end time, pass either a timestamp or a string, eg '2021-09-20T10:20:30' or 'Tue Sep 28 2021 10:58:01'."
    time_zone = None
      .type = str
      .help = "Time zone offset from UTC."
    """
)

sample_scope = freephil.parse(
    """
    sample {
      name = None
        .type = str
        .help = "Descriptive name of the sample."
      temperature = None
        .type = str
        .help = "Sample temperature with units."
    }
    """
)

coord_system_scope = freephil.parse(
    """
    coord_system {
      convention = None
        .type = str
        .help = "Name of the new coordinate system."
      origin = None
        .type = floats
        .help = "Location of the origin e.g. 0,0,0."
      vectors = None
        .type = floats
        .help = "x,y,z axes vectors for the transformation."
    }
    """
)

if __name__ == "__main__":
    print(detector_scope.as_str())
    print(module_scope.as_str())
    print(goniometer_scope.as_str())
    print(instrument_scope.as_str())
    print(timestamp_scope.as_str())
    print(sample_scope.as_str())



"""
Tools to extract goniometer and detector parameters from GDA JSON files.
"""
from __future__ import annotations

import json
from pathlib import Path
from typing import Dict, List, Tuple

from nexgen.nxs_utils import Axis, EigerDetector, TransformationType, TristanDetector
from nexgen.nxs_utils.Detector import DetectorType, UnknownDetectorTypeError
from nexgen.utils import Point3D


class JSONParamsIO:
    """Read JSON file and exctract parameters."""

    def __init__(self, json_file: Path | str):
        self.json_file = json_file
        self.params = self._read_file()

    def _read_file(self) -> Dict:
        with open(self.json_file, "r") as fh:
            params = json.load(fh)
        return params

    def _find_axis_depends_on(self, dep_val: str) -> str:
        depends = dep_val.split("_")[1] if "_" in dep_val else dep_val
        if depends in ["x", "y", "z"]:
            depends = f"sam_{depends}"
        return depends

    def get_coordinate_frame(self) -> str:
        """Get the coordinate frame from geometry json file."""
        return self.params["geometry"]

    def get_goniometer_axes_from_file(self) -> List[Axis]:
        """Read the axes information from the GDA-supplied json file."""
        axes_list = []
        for v in self.params.values():
            if isinstance(v, dict) and v["location"] == "sample":
                ax_depends = self._find_axis_depends_on(v["depends_on"])
                ax_type = (
                    TransformationType.ROTATION
                    if v["type"] == "rotation"
                    else TransformationType.TRANSLATION
                )
                axes_list.append(
                    Axis(v["ds_name"], ax_depends, ax_type, tuple(v["vector"]))
                )
        return axes_list

    def get_detector_axes_from_file(self) -> List[Axis]:
        """Read the detector axes information from the GDA-supplied json file."""
        axes_list = []
        for v in self.params.values():
            if isinstance(v, dict) and v["location"] == "detector":
                ax_type = (
                    TransformationType.ROTATION
                    if v["type"] == "rotation"
                    else TransformationType.TRANSLATION
                )
                axes_list.append(
                    Axis(v["ds_name"], v["depends_on"], ax_type, tuple(v["vector"]))
                )
        return axes_list

    def get_detector_params_from_file(self) -> DetectorType:
        """Read the detector parameters from the GDA-supplied json file."""
        if "tristan" in self.params.keys():
            tristan_params = self.params["tristan"]
            material = (
                "Si"
                if tristan_params["sensor_material"] == "Silicon"
                else tristan_params["sensor_material"]
            )
            thickness = (
                str(tristan_params["sensor_thickness"])
                + tristan_params["sensor_thickness_units"]
            )
            pix = [
                str(i) + tristan_params["pixel_size_units"]
                for i in tristan_params["pixel_size_sf"][::-1]
            ]
            det_params = TristanDetector(
                description=tristan_params["description"],
                image_size=tristan_params["data_size_sf"][::-1],
                sensor_material=material,
                sensor_thickness=thickness,
                pixel_size=pix,
                detector_type=tristan_params["detector_type"],
            )
        elif "eiger" in self.params.keys():
            eiger_params = self.params["eiger"]
            material = eiger_params["sensor_material"]
            pix = [
                str(i) + eiger_params["pixel_size_units"]
                for i in eiger_params["pixel_size"]
            ]
            if material == "Silicon":
                material = "Si"
            det_params = EigerDetector(
                description=eiger_params["description"],
                image_size=eiger_params["size"],
                sensor_material=material,
                overload=50649,
                underload=-1,
                pixel_size=pix,
            )
        else:
            raise UnknownDetectorTypeError("Unknown detector in GDA JSON file.")

        return det_params

    def get_fast_and_slow_direction_vectors_from_file(
        self,
        det_type: str,
    ) -> Tuple[Point3D, Point3D]:
        """Read detector fast and slow axes from the GDA-supplied json file."""
        det_name = "eiger" if "eiger" in det_type.lower() else "tristan"
        det_params = self.params[det_name]
        fast_axis = Point3D(
            x=det_params["fast_dir"][0],
            y=det_params["fast_dir"][1],
            z=det_params["fast_dir"][2],
        )
        slow_axis = Point3D(
            x=det_params["slow_dir"][0],
            y=det_params["slow_dir"][1],
            z=det_params["slow_dir"][2],
        )
        return fast_axis, slow_axis



import numpy as np

# import pytest
from numpy.testing import assert_array_equal

from nexgen.nxs_utils import Axis, Goniometer, TransformationType

axes_list = [
    Axis("omega", ".", TransformationType.ROTATION, (0, 0, -1), 0.0),
    Axis("sam_z", "omega", TransformationType.TRANSLATION, (0, 0, 1), 0.0),
    Axis("sam_y", "sam_z", TransformationType.TRANSLATION, (0, 1, 0), 0.0, 0.1, 10),
    Axis("sam_x", "sam_y", TransformationType.TRANSLATION, (1, 0, 0), 0.0, 0.1, 10),
]


def test_define_scan_from_gonio():
    osc_scan, grid_scan = Goniometer(axes_list).define_scan_from_goniometer_axes()
    assert_array_equal(osc_scan["omega"], np.repeat(0.0, 100))
    assert "sam_x" in list(grid_scan.keys()) and "sam_y" in list(grid_scan.keys())
    assert grid_scan["sam_y"][0] == 0.0
    assert grid_scan["sam_y"][-1] == 0.9
    assert np.min(grid_scan["sam_x"]) == 0.0 and np.max(grid_scan["sam_x"]) == 0.9
    assert round(grid_scan["sam_x"][1] - grid_scan["sam_x"][0], 1) == 0.1


def test_define_scan_given_a_rotation_scan():
    scan = {"omega": np.arange(0, 10, 1)}
    osc_scan, grid_scan = Goniometer(axes_list, scan).define_scan_from_goniometer_axes()
    assert grid_scan is None
    assert_array_equal(osc_scan["omega"], scan["omega"])


def test_define_scan_given_a_grid_scan():
    # For testing pusposes, reset number of images for omega
    axes_list[0].num_steps = 0
    axes_list[0].increment = 0
    # Whatever scan is passed overrides the gonio definition.
    scan = {
        "sam_y": np.array([0, 0, 0, 1, 1, 1]),
        "sam_x": np.array([0, 1, 2, 2, 1, 0]),
    }
    osc_scan, grid_scan = Goniometer(axes_list, scan).define_scan_from_goniometer_axes()
    assert osc_scan is not None
    assert_array_equal(osc_scan["omega"], np.repeat(0.0, 6))
    assert "sam_x" in list(grid_scan.keys()) and "sam_y" in list(grid_scan.keys())
    assert_array_equal(grid_scan["sam_y"], scan["sam_y"])
    assert_array_equal(grid_scan["sam_x"], scan["sam_x"])


def test_get_number_of_scan_points_given_a_scan():
    scan = {
        "sam_y": np.array([0, 0, 0, 1, 1, 1]),
        "sam_x": np.array([0, 1, 2, 2, 1, 0]),
    }
    num_frames = Goniometer(axes_list, scan).get_number_of_scan_points()
    assert num_frames == len(scan["sam_x"])


def test_get_number_of_scan_points():
    num_frames = Goniometer(
        [
            Axis("omega", ".", TransformationType.ROTATION, (0, 0, -1), 90.0),
            Axis("phi", "omega", TransformationType.ROTATION, (0, 0, -1), 0.0, 0.2, 10),
        ],
    ).get_number_of_scan_points()
    assert num_frames == 10


def test_define_scan_axes_for_event_mode_given_scan():
    scan = {"phi": (-90, 0)}
    osc_scan, grid_scan = Goniometer(
        [Axis("phi", ".", TransformationType.ROTATION, (0, 0, -1), 0.0)], scan
    ).define_scan_axes_for_event_mode()
    assert grid_scan is None
    assert_array_equal(osc_scan["phi"], scan["phi"])


def test_define_scan_axes_for_event_mode():
    osc_scan, _ = Goniometer(
        [Axis("phi", ".", TransformationType.ROTATION, (0, 0, -1), 0.0), *axes_list[1:]]
    ).define_scan_axes_for_event_mode()
    assert_array_equal(osc_scan["phi"], (0, 0))


def test_define_scan_axes_for_event_mode_given_new_end_position():
    osc_scan, _ = Goniometer(axes_list).define_scan_axes_for_event_mode(
        end_position=180
    )
    assert_array_equal(osc_scan["omega"], (0, 180))


def test_given_scan_gonio_positions_correctly_updated():
    scan = {
        "sam_y": np.repeat(np.arange(0, 1.0, 0.2), 2),
        "sam_x": np.array(np.tile([1.0, 2.0], 5)),
    }
    _, _ = Goniometer(axes_list, scan).define_scan_from_goniometer_axes()
    Xidx, Yidx = (3, 2)
    assert axes_list[Xidx].start_pos != 0.0
    assert axes_list[Xidx].start_pos == 1.0 and axes_list[Xidx].end_pos == 2.0
    assert axes_list[Yidx].start_pos == 0.0 and axes_list[Yidx].end_pos == 0.8
    assert axes_list[Yidx].increment == 0.2 and axes_list[Xidx].increment == 1.0
    assert axes_list[Xidx].num_steps == 2 and axes_list[Yidx].num_steps == 5


def test_gonio_axes_have_correct_values_if_given_forward_or_reverse_rotation_scan():
    # Forward scan
    scan_fw = {"omega": np.arange(1, 2, 0.1)}
    gonio_fw = Goniometer(axes_list, scan_fw)
    # Omega first acxis in list
    assert gonio_fw.axes_list[0].start_pos == 1.0
    assert gonio_fw.axes_list[0].increment == 0.1
    assert gonio_fw.axes_list[0].num_steps == 10

    # Reverse scan
    scan_rw = {"omega": np.arange(7.5, 7, -0.1)}
    gonio_rw = Goniometer(axes_list, scan_rw)
    assert gonio_rw.axes_list[0].start_pos == 7.5
    assert gonio_rw.axes_list[0].increment == -0.1
    assert gonio_rw.axes_list[0].num_steps == 5



from nexgen.beamlines.beamline_utils import BeamlineAxes, PumpProbe
from nexgen.nxs_utils import Axis, TransformationType
from nexgen.utils import Point3D
from tests.beamlines.conftest import axes_list as gonio_axes


def test_pump_probe():
    pump_probe = PumpProbe()
    pump_probe.pump_status = True
    pump_probe.pump_exposure = 0.01
    assert pump_probe.pump_status is True
    assert pump_probe.pump_exposure == 0.01
    assert pump_probe.pump_delay is None


def test_pump_probe_dict():
    pump_probe = PumpProbe().to_dict()
    assert list(pump_probe.keys()) == [
        "pump_status",
        "pump_exposure",
        "pump_delay",
        "pump_repeat",
    ]
    assert pump_probe["pump_status"] is False


def test_pump_status_set_to_true_if_exposure_is_passed():
    pump_probe = PumpProbe(pump_exposure=0.1)
    assert pump_probe.pump_status is True


def test_beamline_axes():
    bl = BeamlineAxes(
        gonio=gonio_axes,
        det_axes=[Axis("det_z", ".", TransformationType.TRANSLATION, (0, 0, 1))],
        fast_axis=(0, 1, 0),
        slow_axis=(1, 0, 0),
    )
    assert isinstance(bl.fast_axis, Point3D) is True
    assert isinstance(bl.slow_axis, Point3D) is True



import numpy as np
import pytest
from numpy.testing import assert_array_equal

from nexgen.nxs_utils.Axes import Axis, TransformationType
from nexgen.nxs_utils.ScanUtils import (
    ScanAxisError,
    ScanAxisNotFoundError,
    calculate_scan_points,
    identify_grid_scan_axes,
    identify_osc_axis,
)

test_axis_list = [
    Axis("omega", ".", TransformationType.ROTATION, (0, 0, -1), -90),
    Axis("phi", "omega", TransformationType.ROTATION, (0, 0, -1), 180),
    Axis("sam_y", "phi", TransformationType.TRANSLATION, (0, 1, 0), 0, 0.1, 10),
    Axis("sam_x", "sam_y", TransformationType.TRANSLATION, (1, 0, 0), 0, 0.2, 5),
]


def test_identify_osc_axis():
    osc_axis = identify_osc_axis(
        [
            Axis("chi", ".", TransformationType.ROTATION, (0, 0, -1), -90, 0.1, 100),
            Axis("phi", "chi", TransformationType.ROTATION, (0, 0, -1), 180),
        ]
    )
    assert osc_axis == "chi"


def test_identify_osc_axis_with_just_one_rotation_axis():
    osc_axis = identify_osc_axis(
        [Axis("phi", "omega", TransformationType.ROTATION, (0, 0, -1), 180)]
    )
    assert osc_axis == "phi"


def test_identify_osc_axis_from_default_value():
    osc_axis = identify_osc_axis(test_axis_list, default="omega")
    assert osc_axis == "omega"


def test_identify_grid_scan_axis():
    grid_axes = identify_grid_scan_axes(test_axis_list)
    assert "sam_x" in grid_axes
    assert "sam_y" in grid_axes


def test_identify_grid_scan_axis_returns_empty_list_if_no_scan():
    grid_axes = identify_grid_scan_axes(
        [Axis("sam_z", "phi", TransformationType.TRANSLATION, (0, 0, 1), 0)],
    )
    assert grid_axes == []


def test_osc_axis_search_fails_for_multiple_moving_axes():
    with pytest.raises(ScanAxisNotFoundError):
        identify_osc_axis(
            [
                Axis(
                    "chi", ".", TransformationType.ROTATION, (0, 0, -1), -90, 0.1, 100
                ),
                Axis(
                    "phi", "chi", TransformationType.ROTATION, (0, 0, -1), 180, 0.5, 50
                ),
            ]
        )


def test_osc_axis_search_fails_for_no_axes():
    with pytest.raises(ScanAxisNotFoundError):
        identify_osc_axis([])


def test_grid_axis_search_fails_for_no_axes():
    with pytest.raises(ScanAxisNotFoundError):
        identify_grid_scan_axes([])


def test_calculate_scan_points_for_rotation_scan():
    ax = Axis("chi", ".", TransformationType.ROTATION, (0, 0, -1), 0.0, 0.5, 4)
    rot = calculate_scan_points(ax, rotation=True)
    assert "chi" in rot.keys()
    assert len(rot["chi"]) == 4
    assert rot["chi"][0] == ax.start_pos
    assert rot["chi"][-1] == ax.end_pos
    assert_array_equal(rot["chi"], np.array([0.0, 0.5, 1.0, 1.5]))


def test_calculate_scan_points_for_reverse_rotation():
    ax = Axis("phi", ".", TransformationType.ROTATION, (0, 0, -1), 0.0, -0.2, 5)
    rev = calculate_scan_points(ax, rotation=True)
    assert rev[ax.name][0] == 0.0
    assert round(rev[ax.name][-1], 1) == -0.8


def test_calculate_omega_points_for_no_rotation():
    rot = calculate_scan_points(test_axis_list[0], rotation=True, tot_num_imgs=10)
    assert test_axis_list[0].name in rot.keys()
    assert len(rot[test_axis_list[0].name]) == 10
    assert_array_equal(
        rot[test_axis_list[0].name], np.repeat(test_axis_list[0].start_pos, 10)
    )


def test_calculate_scan_points_fails_for_wrong_axis_type():
    with pytest.raises(ScanAxisError):
        calculate_scan_points(test_axis_list[-1], rotation=True)


def test_calculate_scan_points_fails_for_missing_number_of_steps():
    with pytest.raises(ValueError):
        calculate_scan_points(test_axis_list[0], rotation=True)


def test_calculate_scan_points_for_linear_scan():
    lin = calculate_scan_points(test_axis_list[-1])
    assert len(lin) == 1
    assert "sam_x" in lin.keys()
    assert len(lin["sam_x"]) == test_axis_list[-1].num_steps
    assert round(lin["sam_x"][1] - lin["sam_x"][0], 1) == test_axis_list[-1].increment


def test_calculate_scan_points_for_snaked_grid_scan():
    grid = calculate_scan_points(
        test_axis_list[-2],
        test_axis_list[-1],
    )
    assert "sam_x" in grid.keys() and "sam_y" in grid.keys()
    assert len(grid["sam_x"]) == len(grid["sam_y"])
    assert round(grid["sam_x"][1] - grid["sam_x"][0], 1) == 0.2
    assert grid["sam_x"][-1] == test_axis_list[-1].start_pos


def test_calculate_scan_points_for_non_snaked_grid_scan():
    grid = calculate_scan_points(
        test_axis_list[-2],
        test_axis_list[-1],
        snaked=False,
    )
    assert "sam_x" in grid.keys() and "sam_y" in grid.keys()
    assert (
        len(grid["sam_x"])
        == test_axis_list[-2].num_steps * test_axis_list[-1].num_steps
    )
    assert round(grid["sam_y"][5] - grid["sam_y"][0], 1) == 0.1
    assert grid["sam_x"][-1] == test_axis_list[-1].end_pos



"""
Utilities for writing new NeXus format files.
"""

from __future__ import annotations

import logging
import math
from datetime import datetime, timedelta
from pathlib import Path
from typing import List, Literal, Tuple

import h5py  # isort: skip
import numpy as np
from hdf5plugin import Bitshuffle, Blosc
from numpy.typing import ArrayLike

# Logger
NXclassUtils_logger = logging.getLogger("nexgen.NXclass_writers.utils")
NXclassUtils_logger.setLevel(logging.DEBUG)

# Define Timestamp dataset names
TSdset = Literal["start_time", "end_time", "end_time_estimated"]


def create_attributes(nxs_obj: h5py.Group | h5py.Dataset, names: Tuple, values: Tuple):
    """
    Create or overwrite attributes with additional metadata information.

    Args:
        nxs_obj (h5py.Group | h5py.Dataset): NeXus object to which the attributes should be attached.
        names (Tuple): The names of the new attributes.
        values (Tuple): The attribute values asociated to the names.
    """
    for n, v in zip(names, values):
        if isinstance(v, str):
            # If a string, convert to numpy.string_
            v = np.string_(v)
        h5py.AttributeManager.create(nxs_obj, name=n, data=v)


def set_dependency(dep_info: str, path: str = None):
    """
    Define value for "depends_on" attribute.
    If the attribute points to the head of the dependency chain, simply pass "." for dep_info.

    Args:
        dep_info (str): The name of the transformation upon which the current one depends on.
        path (str): Where the transformation is. Set to None, if passed it points to location in the NeXus tree.
    Returns:
        The value to be passed to the attribute "depends_on"
    """
    if dep_info == ".":
        return np.string_(".")
    if path:
        if path.endswith("/") is False:
            path += "/"
        return np.string_(path + dep_info)
    else:
        return np.string_(dep_info)


def calculate_origin(
    beam_center_fs: List | Tuple,
    fs_pixel_size: List | Tuple,
    fast_axis_vector: Tuple,
    slow_axis_vector: Tuple,
    mode: str = "1",
) -> Tuple[List, float]:
    """
    Calculate the offset of the detector.

    This function returns the detector origin array, which is saved as the vector attribute of the module_offset field.
    The value to set the module_offset to is also returned: the magnitude of the displacement if the vector is normalized, 1.0 otherwise
    Assumes that fast and slow axis vectors have already been converted to mcstas if needed.

    Args:
        beam_center_fs (List | Tuple): Beam center position in fast and slow direction.
        fs_pixel_size (List | Tuple): Pixel size in fast and slow direction, in m.
        fast_axis_vector (Tuple): Fast axis vector.
        slow_axis_vector (Tuple): Slow axis vector.
        mode (str, optional): Decide how origin should be calculated.
                            If set to "1" the displacement vector is un-normalized and the offset value set to 1.0.
                            If set to "2" the displacement is normalized and the offset value is set to the magnitude of the displacement.
                            Defaults to "1".

    Returns:
        det_origin (List): Displacement of beam center, vector attribute of module_offset.
        offset_val (float): Value to assign to module_offset, depending whether det_origin is normalized or not.
    """
    # what was calculate module_offset
    x_scaled = beam_center_fs[0] * fs_pixel_size[0]
    y_scaled = beam_center_fs[1] * fs_pixel_size[1]
    # Detector origin
    det_origin = x_scaled * np.array(fast_axis_vector) + y_scaled * np.array(
        slow_axis_vector
    )
    det_origin = list(-det_origin)
    if mode == "1":
        offset_val = 1.0
    else:
        offset_val = math.hypot(*det_origin[:-1])
    return det_origin, offset_val


def find_number_of_images(datafile_list: List[Path], entry_key: str = "data") -> int:
    """
    Calculate total number of images when there's more than one input HDF5 file.

    Args:
        datafile_list (List[Path]): List of paths to the input image files.
        entry_key (str):    Key for the location of the images inside the data files. Defaults to "data".

    Returns:
        num_images (int): Total number of images.
    """
    num_images = 0
    for filename in datafile_list:
        with h5py.File(filename, "r") as f:
            num_images += f[entry_key].shape[0]
    return int(num_images)


def calculate_estimated_end_time(
    start_time: datetime | str, tot_collection_time: float
) -> str:
    time_format = r"%Y-%m-%dT%H:%M:%SZ"

    if isinstance(start_time, str):
        start_time = start_time.format("%Y-%m-%dT%H:%M:%S")
        start_time = datetime.strptime(start_time.strip("Z"), time_format.strip("Z"))

    est_end = start_time + timedelta(seconds=tot_collection_time)
    return est_end.strftime(time_format)


def mask_and_flatfield_writer(
    nxdet_grp: h5py.Group,
    dset_name: str,
    dset_data: str | ArrayLike,
    applied_val: bool,
):
    """ Utility function to write mask or flatfield to NXdetector group for image data when not \
    already linked to the _meta.h5 file.
    If the pixel_mask/flatfield data is passed as a string, it will be assumed to be a file path and \
    the writer will try to set up an external link to it.

    Args:
        nxdet_grp (h5py.Group): Handle to HDF5 NXdetector group.
        dset_name (str): Name of the new field/dataset to be written.
        dset_data (str | ArrayLike): Dataset data to be written in the field. Can be a string or an \
        array-like dataset. If the data type is a numpy ndarray, it will be compressed before writing.
        applied_val (bool): Value to write to the `{flatfield,pixel_mask}_applied` fields.
    """
    if dset_data is None:
        NXclassUtils_logger.warning(
            f"""
            No copy of the {dset_name} has been found, either as a file or dataset.
            Fields {dset_name} and {dset_name}_applied will not be written to file.
            """
        )
        return
    nxdet_grp.create_dataset(
        f"{dset_name}_applied",
        data=applied_val,
    )
    NXclassUtils_logger.debug(f"{dset_name}_applied set to: {applied_val}.")
    if isinstance(dset_data, str):
        try:
            link_path = Path(dset_data)
            NXclassUtils_logger.debug(
                f"Setting external link for {dset_name} to {link_path}."
            )
            nxdet_grp[dset_name] = h5py.ExternalLink(link_path.name, "/")
        except Exception as e:
            NXclassUtils_logger.error(
                f"Impossible to write external link to {dset_data} for {dset_name}."
                "Field {dset_name} not written."
            )
            NXclassUtils_logger.error(f"{e}", exc_info=1)
    elif isinstance(dset_data, np.ndarray):
        NXclassUtils_logger.debug(f"Writing a compressed copy of array in {dset_name}.")
        write_compressed_copy(nxdet_grp, dset_name, data=dset_data)
    else:
        NXclassUtils_logger.debug(
            f"{dset_name} of type {type(dset_data)}, writing as is."
        )
        nxdet_grp.create_dataset(dset_name, data=dset_data)
    return


def mask_and_flatfield_writer_for_event_data(
    nxdet_grp: h5py.Group,
    dset_name: str,
    dset_data_file: str,
    applied_val: bool,
    wdir: Path,
    detector_name: str = "tristan",
):
    if dset_data_file is None:
        NXclassUtils_logger.warning(
            f"No {dset_name} data file passed; {dset_name} won't be written."
        )
        return

    nxdet_grp.create_dataset(f"{dset_name}_applied", data=applied_val)
    NXclassUtils_logger.info(f"Looking for file {dset_data_file} in {wdir.as_posix()}.")
    filename = [
        wdir / dset_data_file for f in wdir.iterdir() if dset_data_file == f.name
    ]
    if filename:
        NXclassUtils_logger.info(f"File {dset_name} found in working directory.")
        write_compressed_copy(
            nxdet_grp,
            dset_name,
            filename=filename[0],
            filter_choice="blosc",
            dset_key="image",
        )
    else:
        NXclassUtils_logger.warning(
            f"No {dset_name} file found in working directory."
            "Writing an ExternalLink."
        )
        file_loc = Path(dset_data_file)
        image_key = "image" if "tristan" in detector_name.lower() else "/"
        nxdet_grp[dset_name] = h5py.ExternalLink(file_loc.name, image_key)
    return


# Copy and compress a dataset inside a specified NXclass
def write_compressed_copy(
    nxgroup: h5py.Group,
    dset_name: str,
    data: ArrayLike = None,
    filename: Path | str = None,
    filter_choice: str = "bitshuffle",
    dset_key: str = "image",
    **kwargs,
):
    """
    Write a compressed copy of some dataset in the desired HDF5 group, using the filter of choice with lz4 compression. Available filters \
    at this time include "Blosc" and "Bitshuffle" (default).
    The main application for this function in nexgen is to write a compressed copy of a pixel mask or a flatfield file/dataset \
    directly into the NXdetector group of a NXmx NeXus file.
    The data and filename arguments are mutually exclusive as only one of them can be used as input.
    If a filename is passed, it is also required to pass the key for the relevant dataset to be copied. Failure to do so will result \
    in nothing being written to the NeXus file.

    Args:
        nxgroup (h5py.Group): Handle to HDF5 group.
        dset_name (str): Name of the new dataset to be written.
        data (ArrayLike, optional): Dataset to be compressed. Defaults to None.
        filename (Path | str, optional): Filename containing the dataset to be compressed into the NeXus file. Defaults to None.
        filter_choice (str, optional): Filter to be used for compression. Either blosc or bitshuffle. Defaults to bitshuffle.
        dset_key (str, optional): Dataset name inside the passed file. Defaults to "image".

    Keyword Args:
        block_size (int, optional): Number of elements per block, it needs to be divisible by 8. Needed for Bitshuffle filter. \
            Defaults to 0.

    Raises:
        ValueError: If both a dataset and a filename have been passed to the function.
    """
    if data is not None and filename is not None:
        raise ValueError(
            "The dset and filename arguments are mutually exclusive."
            "Please pass only the one from which the data should be copied."
        )
    if filename and not dset_key:
        NXclassUtils_logger.warning(
            f"Missing key to find the dataset to be copied inside {filename}. {dset_name} will not be written into the NeXus file."
        )
        return

    if filename:
        with h5py.File(filename, "r") as fh:
            data = fh[dset_key][()]

    if filter_choice.lower() == "blosc":
        nxgroup.create_dataset(
            dset_name, data=data, **Blosc(cname="lz4", shuffle=Blosc.BITSHUFFLE)
        )
    elif filter_choice.lower() == "bitshuffle":
        block_size = (
            0 if "block_size" not in list(kwargs.keys()) else kwargs["block_size"]
        )
        nxgroup.create_dataset(
            dset_name, data=data, **Bitshuffle(nelems=block_size, cname="lz4")
        )
    else:
        NXclassUtils_logger.warning(
            "Unknown filter choice, no dataset will be written."
        )
        return
    NXclassUtils_logger.info(
        f"A compressed copy of the {dset_name} has been written into the NeXus file."
    )



import numpy as np
import pytest
from numpy.testing import assert_array_equal

from nexgen.beamlines.beamline_utils import PumpProbe
from nexgen.beamlines.SSX_expt import run_extruder, run_fixed_target
from tests.beamlines.conftest import axes_list

test_pump = PumpProbe()
test_pump.pump_status = True
test_pump.pump_exposure = 0.1

test_chip_dict = {
    "X_NUM_STEPS": [0, 20],
    "Y_NUM_STEPS": [0, 20],
    "X_STEP_SIZE": [0, 0.125],
    "Y_STEP_SIZE": [0, 0.125],
    "X_START": [0, 0],
    "Y_START": [0, 0],
    "Z_START": [0, 0],
    "X_NUM_BLOCKS": [0, 2],
    "Y_NUM_BLOCKS": [0, 2],
    "X_BLOCK_SIZE": [0, 3.175],
    "Y_BLOCK_SIZE": [0, 3.175],
    "N_EXPOSURES": [0, 1],
    "PUMP_REPEAT": [0, 0],
}


def test_run_extruder():
    gonio, osc, info = run_extruder(axes_list, 10, test_pump, "omega")
    idx = [n for n, ax in enumerate(axes_list) if ax.name == "omega"][0]
    assert gonio[idx].start_pos == 0.0
    assert gonio[idx].increment == 0.0
    assert gonio[idx].num_steps == 10
    assert list(osc.keys()) == ["omega"]
    assert_array_equal(osc["omega"], np.zeros(10))
    assert isinstance(info, dict)
    assert info["pump_exposure"] == 0.1 and info["pump_delay"] is None


def test_run_extruder_with_non_zero_omega():
    axes_list[0].start_pos = 90.0
    gonio, osc, _ = run_extruder(axes_list, 10, test_pump)
    assert gonio[0].start_pos == 90.0
    assert gonio[0].increment == 0.0
    assert_array_equal(osc["omega"], np.repeat(90.0, 10))


def test_run_fixed_target(dummy_chipmap_file_single_block):
    transl, info = run_fixed_target(
        axes_list,
        test_chip_dict,
        dummy_chipmap_file_single_block.name,
        test_pump,
    )
    assert list(transl.keys()) == ["sam_y", "sam_x"]
    assert len(transl["sam_x"]) == len(transl["sam_y"])
    assert len(transl["sam_y"]) == 400
    assert info["n_exposures"] == 1


def test_run_fixed_target_with_wrong_axis_in_list(dummy_chipmap_file_single_block):
    with pytest.raises(ValueError):
        _ = run_fixed_target(
            axes_list,
            test_chip_dict,
            dummy_chipmap_file_single_block.name,
            test_pump,
            scan_axes=["phi", "sam_x"],
        )


def test_run_fixed_target_with_missing_scan_axis(dummy_chipmap_file_single_block):
    with pytest.raises(IndexError):
        _ = run_fixed_target(
            axes_list,
            test_chip_dict,
            dummy_chipmap_file_single_block.name,
            test_pump,
            scan_axes=["sam_z"],
        )


def test_fixed_target_raises_error_if_no_chip_info(dummy_chipmap_file_single_block):
    with pytest.raises(ValueError):
        _ = run_fixed_target(
            axes_list,
            {},
            dummy_chipmap_file_single_block.name,
            test_pump,
        )


def test_fixed_target_for_multiple_exposures(dummy_chipmap_file_single_block):
    test_chip_dict["N_EXPOSURES"] = [0, "2"]
    transl, info = run_fixed_target(
        axes_list,
        test_chip_dict,
        dummy_chipmap_file_single_block.name,
        test_pump,
    )
    assert list(transl.keys()) == ["sam_y", "sam_x"]
    assert len(transl["sam_x"]) == len(transl["sam_y"])
    assert len(transl["sam_y"]) == 800
    assert info["n_exposures"] == 2


def test_fixed_target_with_upwards_blocks(dummy_chipmap_file_multi_block):
    test_chip_dict["N_EXPOSURES"] = [0, "1"]
    transl, info = run_fixed_target(
        axes_list,
        test_chip_dict,
        dummy_chipmap_file_multi_block.name,
        test_pump,
    )
    assert list(transl.keys()) == ["sam_y", "sam_x"]
    assert len(transl["sam_x"]) == len(transl["sam_y"])
    assert len(transl["sam_y"]) == 1600
    assert info["n_exposures"] == 1


def test_fixed_target_fullchip_with_multiple_exposures(dummy_chipmap_file_multi_block):
    test_chip_dict["N_EXPOSURES"] = [0, "2"]
    transl, info = run_fixed_target(
        axes_list,
        test_chip_dict,
        dummy_chipmap_file_multi_block.name,
        test_pump,
    )
    assert list(transl.keys()) == ["sam_y", "sam_x"]
    assert len(transl["sam_x"]) == len(transl["sam_y"])
    assert len(transl["sam_y"]) == 3200
    assert info["n_exposures"] == 2



"""
Create a nexus file for electron diffraction collections using singla detector.
"""

from __future__ import annotations

import logging
from pathlib import Path
from typing import Dict, List

from .. import log
from ..nxs_utils import Attenuator, Beam, Detector, Goniometer, SinglaDetector
from ..nxs_utils.ScanUtils import calculate_scan_points
from ..nxs_write.NXmxWriter import EDNXmxFileWriter
from ..nxs_write.write_utils import find_number_of_images
from ..tools.ED_tools import (
    extract_detector_info_from_master,
    extract_exposure_time_from_master,
    extract_start_time_from_master,
    find_beam_centre,
)
from ..utils import coerce_to_path, find_in_dict, get_iso_timestamp, get_nexus_filename
from .ED_params import ED_coord_system, EDSingla, EDSource

logger = logging.getLogger("nexgen.EDNeXusWriter")

# Define entry_key if dealing with singla detector
SINGLA_DATA_ENTRY_KEY = "/entry/data/data"


def singla_nexus_writer(
    master_file: Path | str,
    det_distance: float,
    exp_time: float,
    ED_cs: Dict = ED_coord_system,
    datafiles: List[Path | str] = None,
    convert2mcstas: bool = False,
    **params,
):
    """Gather all collection parameters and write the NeXus file for an electron \
    diffraction collectio using SINGLA detector.

    Args:
        master_file (Path | str): Singla master file.
        det_distance (float): Sample-detector distance, in mm.
        exp_time (float): Exposure time, in s.
        ED_cs (Dict, optional): Definition of the ED coordinate system in use. Defaults to\
            {"convention": "ED",
            "origin": (0, 0, 0),
            "x": Axis("x", ".", "translation", [0, 1, 0]),
            "y":  Axis("y", "x", "translation", [-1, 0, 0]),
            "z":  Axis("z", "y", "translation", [0, 0, 1]),}
        datafiles (List[Path | str], optional): List of data files. Defaults to None.
        convert2mcstas (bool, optional): Convert vectors to mcstas if required. \
            Defaults to False.

    Keyword Args:
        n_imgs (int): Total number of images in collection.
        scan_axis (List[str, float, float]): Rotation axis name, start and increment.
        outdir (Path | str): Directory where to save the file. Only specify if different \
            from meta_file directory.
        beam_center (List[float, float]): Beam center position, in pixels.
        wavelength (float): Wavelength of incident beam, in A.
        start_time (datetime): Experiment start time.
        new_source_info (Dict): Information about Source that might differ from the default.\
            eg. {"facility_id": "MICROSCOPE", "name": "Not Diamond"}
        vds_writer (str): Write dataset or external file.
    """
    master_file = coerce_to_path(master_file)

    # Get NeXus filename
    nxsfile = get_nexus_filename(master_file)
    if find_in_dict("outdir", params) and params["outdir"]:
        wdir = coerce_to_path(params["outdir"])
        # Reset the location of the NeXus file
        nxsfile = wdir / nxsfile.name
    else:
        wdir = master_file.parent

    # Set up logging config
    logfile = wdir / "EDnxs.log"
    log.config(logfile.as_posix())

    logger.info("NeXus file writer for electron diffraction data from Singla.")
    logger.info(f"Collection directory: {master_file.parent}.")
    logger.info("NeXus file will be saved as %s" % nxsfile.name)

    # Data files
    if not datafiles:
        datafiles = [
            f
            for f in master_file.parent.iterdir()
            if nxsfile.stem + "_data" in f.as_posix()
        ]
    logger.info(f"{len(datafiles)} data files in directory.")

    # Total number of images
    if find_in_dict("n_imgs", params) and params["n_imgs"]:
        tot_num_imgs = params["n_imgs"]
    else:
        tot_num_imgs = find_number_of_images(datafiles, SINGLA_DATA_ENTRY_KEY)
    logger.info(f"Total number of images: {tot_num_imgs}.")

    # Get start_time timestamp in ISOformat
    if find_in_dict("start_time", params):
        start_time = get_iso_timestamp(params["start_time"])
    else:
        start_time = extract_start_time_from_master(master_file)

    # Update source if new info passed
    source = EDSource
    if find_in_dict("new_source_info", params):
        logger.warning("Updating source information.")
        for k, v in params["new_source_info"].items():
            source.__setattr__(k, v)
            logger.info(f"Source {k} now set to {v}.")
    logger.info(source.__repr__())

    # Define beam and attenuator
    attenuator = Attenuator(transmission=None)
    wl = params["wavelength"] if find_in_dict("wavelength", params) else None
    if not wl:
        logger.warning("Wavelength value was not set, it will default to 0.02 A.")
        wl = 0.02
    beam = Beam(wl)
    logger.info(f"Attenuation: {attenuator.transmission}")
    logger.info(f"Incident beam wavelength: {beam.wavelength}")

    # Define Singla detector
    det_params = SinglaDetector("Dectris Singla 1M", [1062, 1028])
    # Update detector params with info from master file
    logger.info(
        "Looking through Dectris master file to extract at least mask and flatfield."
    )
    det_info = extract_detector_info_from_master(master_file)
    det_params.constants.update(det_info)
    # If beam_centre not passed, define it
    if not find_in_dict("beam_center", params) or params["beam_center"] is None:
        beam_center = find_beam_centre(master_file, datafiles[0])
        logger.info(f"Calculated beam centre to be {beam_center}.")
        if beam_center is None:
            beam_center = (0, 0)
            logger.warning(
                f"Unable to calculate beam centre. It has been set to {beam_center}."
            )
    else:
        beam_center = params["beam_center"]

    # Detector/module axes
    det_axes = EDSingla.det_axes
    det_axes[0].start_pos = det_distance

    if not exp_time:
        logger.warning("Exposure time not set, trying to read it from the master file.")
        exp_time = extract_exposure_time_from_master(master_file)

    if not exp_time:
        raise ValueError(
            "Exposure time not provided. No 'count_time' in the master file."
        )

    # Define detector
    detector = Detector(
        det_params,
        det_axes,
        beam_center,
        exp_time,
        [EDSingla.fast_axis, EDSingla.slow_axis],
    )
    logger.info(detector.__repr__())

    # Goniometer
    gonio_axes = EDSingla.gonio
    if find_in_dict("scan_axis", params):
        gonio_axes[0].name = params["scan_axis"][0]
        gonio_axes[1].depends = params["scan_axis"][0]
        gonio_axes[0].start_pos = params["scan_axis"][1]
        gonio_axes[0].increment = params["scan_axis"][2]
        gonio_axes[0].num_steps = tot_num_imgs
    else:
        gonio_axes[0].num_steps = tot_num_imgs
    # No grid scan, can be added if needed at later time
    OSC = calculate_scan_points(
        gonio_axes[0],
        rotation=True,
        tot_num_imgs=tot_num_imgs,
    )
    logger.info(f"Rotation scan axis: {list(OSC.keys())[0]}.")
    logger.info(
        f"Scan from {list(OSC.values())[0][0]} to {list(OSC.values())[0][-1]}.\n"
    )

    goniometer = Goniometer(gonio_axes, OSC)
    logger.info(goniometer.__repr__())

    vds_writer = (
        "dataset" if not find_in_dict("vds_writer", params) else params["vds_writer"]
    )

    # Start writing
    logger.info("Start writing NeXus file ...")
    try:
        EDFileWriter = EDNXmxFileWriter(
            nxsfile,
            goniometer,
            detector,
            source,
            beam,
            attenuator,
            tot_num_imgs,
            ED_cs,
            convert_to_mcstas=convert2mcstas,
        )
        EDFileWriter.write(datafiles, SINGLA_DATA_ENTRY_KEY, start_time)
        if vds_writer:
            EDFileWriter.write_vds(
                writer_type=vds_writer,
                data_entry_key=SINGLA_DATA_ENTRY_KEY,
                datafiles=datafiles,
            )
        else:
            logger.info("VDS won't be written.")
        logger.info("NeXus file written correctly.")
    except Exception as err:
        logger.exception(err)
        logger.info(f"An error occurred and {nxsfile} couldn't be written correctly.")



# Eiger specific
eiger_modules = {"1M": (1, 2), "4M": (2, 4), "9M": (3, 6), "16M": (4, 8)}
eiger_mod_size = (512, 1028)
eiger_gap_size = (38, 12)
intra_mod_gap = 2

# Tristan specific
clock_freq = int(6.4e8)
tristan_modules = {"10M": (2, 5), "2M": (1, 2)}
tristan_mod_size = (515, 2069)  # (H, V)
tristan_gap_size = (117, 45)

# Pre-defined chunk size
tristan_chunk = 2097152

# Junfrau 1M specific
jungfrau_modules = {"1M": (1, 2)}
jungfrau_mod_size = (514, 1030)  # (slow, fast)
jungfrau_gap_size = (38, 12)
jungfrau_fill_value = 0b10000000000000000000000000000000



"""
Object definition for detectors.
"""
from __future__ import annotations

from dataclasses import dataclass, field
from typing import Dict, List, Literal, Tuple, Union

from dataclasses_json import DataClassJsonMixin

from ..utils import Point3D
from .Axes import Axis


class UnknownDetectorTypeError(Exception):
    pass


EIGER_CONST = {
    "flatfield": "flatfield",
    "flatfield_applied": "_dectris/flatfield_correction_applied",
    "pixel_mask": "mask",
    "pixel_mask_applied": "_dectris/pixel_mask_applied",
    "bit_depth_readout": "_dectris/bit_depth_image",  # Gorilla to be consistent with NeXus format (and make DIALS work)
    "bit_depth_image": "_dectris/bit_depth_image",
    "detector_readout_time": "_dectris/detector_readout_time",
    "threshold_energy": "_dectris/threshold_energy",
    "software_version": "_dectris/software_version",
    "serial_number": "_dectris/detector_number",
}

TRISTAN_CONST = {
    "flatfield": "Tristan10M_flat_field_coeff_with_Mo_17.479keV.h5",
    "flatfield_applied": False,
    "pixel_mask": "Tristan10M_mask_with_spec.h5",
    "pixel_mask_applied": False,
    "software_version": "1.1.3",
    "detector_tick": "1562.5ps",
    "detector_frequency": "6.4e+08Hz",
    "timeslice_rollover": 18,
}

JUNGFRAU_CONST = {
    "flatfield": None,
    "flatfield_applied": False,
    "pixel_mask": None,
    "pixel_mask_applied": False,
    "software_version": "0.0.0",
}

SINGLA_CONST = {
    "flatfield": None,
    "flatfield_applied": False,
    "pixel_mask": None,
    "pixel_mask_applied": False,
    "software_version": "0.0.0",
}


@dataclass
class EigerDetector(DataClassJsonMixin):
    """Define a Dectris Eiger detector.

    Attributes:
        description (str): Detector description.
        image_size (List | Tuple): Dimensions in pixels, passed in the order (slow, fast) axis.
        sensor_material (str): Either Si or CdTe, on the material depends the sensor_thickness.
        overload (int): Saturation value for the detector, data is invalid above this value.
        underload (int): Lowest value measurable for the detector, data is invalid below this value.
        pixel_size (List[str], optional): Size of each detector pixel in both directions, order should be (x, y). Defaults to a pixel size of ['0.075mm', '0.075mm']
        detector_type (str, optional): Description of type of detector. Defaults to 'Pixel'.

    Properties:
        sensor_thickness (str): Defined depending on the sensor material: 0.450mm for Si, 0.750mm CdTe.
        constants (Dict): Dictionary of meta file locations to create the external links to fields such as pixel_mask, flatfield and bit_depth_readout.
    """

    description: str
    image_size: List[int] | Tuple[int, int]
    sensor_material: Literal["Si", "CdTe"]
    overload: int
    underload: int
    pixel_size: List[str | float] = field(
        default_factory=lambda: ["0.075mm", "0.075mm"]
    )
    detector_type: str = "Pixel"

    @property
    def sensor_thickness(self) -> str:
        if self.sensor_material == "Si":
            return "0.450mm"
        else:
            return "0.750mm"

    @property
    def constants(self) -> Dict:
        return EIGER_CONST

    @property
    def hasMeta(self) -> bool:
        return True


@dataclass
class TristanDetector(DataClassJsonMixin):
    """Define a Tristan detector.

    Attributes:
        description (str): Detector description.
        image_size (List | Tuple): Dimensions in pixels, passed in the order (slow, fast) axis.
        sensor_material (str): Sensor material. Defaults to Si.
        sensor_thickness (str): Sensor thickness. Defaults to 0.5mm
        pixel_size (List[str], optional): Size of each detector pixel in both directions, order should be (x, y). Defaults to a pixel size of ['5.5e-05m', '5.5e-05m']
        detector_type (str, optional): Description of type of detector. Defaults to 'Pixel'.
        mode (str): Acquisition mode for Tristan, either images or events. Defaults to events.

    Properties:
        constants (Dict): Detector specific constants, such as locations of pixel_mask and flatfield files, and detector tick, frequency and timeslice rollover for event mode.
    """

    description: str
    image_size: List[int] | Tuple[int, int]
    sensor_material: str = "Si"
    sensor_thickness: str = "0.5mm"
    pixel_size: List[str | float] = field(
        default_factory=lambda: ["5.5e-05m", "5.5e-05m"]
    )
    detector_type: str = "Pixel"
    mode: Literal["events", "images"] = "events"

    @property
    def constants(self) -> Dict:
        return TRISTAN_CONST

    @property
    def hasMeta(self) -> bool:
        return True


@dataclass
class SinglaDetector(DataClassJsonMixin):
    """Define a Dectris Singla detector.

    Attributes:
        description (str): Detector description.
        image_size (List | Tuple): Dimensions in pixels, passed in the order (slow, fast) axis.
        sensor_material (str): Sensor material. Defaults to Si.
        sensor_thickness (str): Sensor thickness. Defaults to 0.450mm
        overload (int): Saturation value for the detector, data is invalid above this value. Defaults to 199996.
        underload (int): Lowest value measurable for the detector, data is invalid below this value. Defaults to -1.
        pixel_size (List[str], optional): Size of each detector pixel in both directions, order should be (x, y). Defaults to a pixel size of ['0.075mm', '0.075mm']
        detector_type (str, optional): Description of type of detector. Defaults to 'HPC'.

    Properties:
        constants (Dict): Dictionary of meta file locations to create the external links to fields such as pixel_mask, flatfield and bit_depth_readout.
    """

    description: str
    image_size: List[int] | Tuple[int, int]
    sensor_material: str = "Si"
    sensor_thickness: str = "0.450mm"
    overload: int = 199996
    underload: int = -1
    pixel_size: List[str | float] = field(
        default_factory=lambda: ["0.075mm", "0.075mm"]
    )
    detector_type: str = "HPC"

    @property
    def constants(self) -> Dict:
        return SINGLA_CONST

    @property
    def hasMeta(self) -> bool:
        return False


@dataclass
class JungfrauDetector(DataClassJsonMixin):
    """Define a Dectris Jungfrau detector.

    Attributes:
        description (str): Detector description.
        image_size (List | Tuple): Dimensions in pixels, passed in the order (slow, fast) axis.
        sensor_material (str): Sensor material. Defaults to Si.
        sensor_thickness (str): Sensor thickness. Defaults to 0.320mm
        overload (int): Saturation value for the detector, data is invalid above this value. Defaults to 1000000.
        underload (int): Lowest value measurable for the detector, data is invalid below this value. Defaults to -10.
        pixel_size (List[str], optional): Size of each detector pixel in both directions, order should be (x, y). Defaults to a pixel size of ['0.075mm', '0.075mm']
        detector_type (str, optional): Description of type of detector. Defaults to 'Pixel'.

    Properties:
        constants (Dict): Dictionary of meta file locations to create the external links to fields such as pixel_mask, flatfield and bit_depth_readout.
    """

    description: str
    image_size: List[int] | Tuple[int, int]
    sensor_material: str = "Si"
    sensor_thickness: str = "0.320mm"
    overload: int = 1000000
    underload: int = -10
    pixel_size: List[str | float] = field(
        default_factory=lambda: ["0.075mm", "0.075mm"]
    )
    detector_type: str = "Pixel"

    @property
    def constants(self) -> Dict:
        return JUNGFRAU_CONST

    @property
    def hasMeta(self) -> bool:
        return False


DetectorType = Union[EigerDetector, TristanDetector, SinglaDetector, JungfrauDetector]


@dataclass
class DetectorModule(DataClassJsonMixin):
    """A class to define the axes of a detector module.

    Attributes:
        fast_axis (Tuple | Point3D): Vector defining the fast_axis direction.
        slow_axis (Tuple | Point3D): Vector defining the slow_axis direction.
    """

    fast_axis: Tuple[float, float, float] | Point3D
    slow_axis: Tuple[float, float, float] | Point3D
    module_offset: str = "1"

    def __post_init__(self):
        if isinstance(self.fast_axis, Point3D):
            self.fast_axis = (self.fast_axis.x, self.fast_axis.y, self.fast_axis.z)
        if isinstance(self.slow_axis, Point3D):
            self.slow_axis = (self.slow_axis.x, self.slow_axis.y, self.slow_axis.z)


class Detector:
    """Detector definition.

    Attributes:
        detector_params: The detector parameters, unique to each detector type.
        detector_axes: The axes where the detector lays, their start positions and vectors in mcstas coordinates.
        beam_center: The beam center position, in pixels.
        exp_time: The collection exposure time, in seconds.
        module: The detector module definition, with fast_axis and slow_axis directions, in mcstas.
    """

    def __init__(
        self,
        detector_params: DetectorType,
        detector_axes: List[Axis],
        beam_center: List[float],
        exposure_time: float,
        module_vectors: List[Point3D] | List[Tuple[float, float, float]],
    ):
        """
        Args:
            detector_params (DetectorType): Specific parameters relative to detector in use eg. TristanDetector("Tristan", [100, 200])
            detector_axes (List[Axes]): List of detector axes.
            beam_center (List[float]): Beam center position, in pixels.
            exposure_time (float): Exposure time of each image/collection, in s.
            module_vectors (List): List of detector module vectors in the order: (fast_axis, slow_axis).
        """
        self.detector_params = detector_params
        self.detector_axes = detector_axes
        self.beam_center = beam_center
        self.exp_time = exposure_time
        self.module = DetectorModule(module_vectors[0], module_vectors[1])

    def __repr__(self) -> str:
        det_msg = (
            f"{self.detector_params.description} \n\t"
            f"Image size {self.detector_params.image_size} pixels of size {self.detector_params.pixel_size}; \n\t"
            f"{self.detector_params.sensor_material} sensor x {self.detector_params.sensor_thickness}; \n\t"
            "Detector axes: \n\t"
        )
        for ax in self.detector_axes:
            det_msg += f"{ax.name}: {ax.start_pos} => {ax.transformation_type.value} on {ax.depends} \n\t"
        det_msg += (
            "Detector module axes: \n\t"
            f"Fast axis: {self.module.fast_axis} \n\t Slow axis: {self.module.slow_axis}\n"
        )
        return f"Detector description: {det_msg}"

    def get_detector_description(self) -> str:
        """Get detector description string."""
        return self.detector_params.description

    def get_detector_mode(self) -> str:
        """Data type collected by the detector.
        If no mode specified in detector parameters, defaults to images.
        """
        if "mode" in self.detector_params.__dataclass_fields__:
            return self.detector_params.mode
        return "images"

    def get_module_info(self):
        """Write the module information to a dictionary."""
        return self.module.to_dict()



import tempfile
from datetime import datetime

import h5py
import numpy as np
import pytest
from numpy.testing import assert_array_equal

from nexgen.tools.DataWriter import build_an_eiger
from nexgen.tools.ED_tools import (
    SinglaMaster,
    centroid_max,
    extract_detector_info_from_master,
    extract_exposure_time_from_master,
    extract_start_time_from_master,
    find_beam_centre,
)

test_goniometer = {
    "axes": ["alpha", "sam_z"],
    "vectors": [1, 0, 0, 0, 0, 1],
    "offsets": [(0, 0, 0), (0, 0, 0)],
}
test_detector = {"axes": ["det_z"], "vectors": [0, 0, 1]}
test_module = {
    "fast_axis": [1, 0, 0],
    "slow_axis": [0, 1, 0],
    "offsets": [0, 0, 0, 0, 0, 0],
}

test_new_coords = {
    "convention": "rotate",
    "origin": (0, 0, 0),
    "x": (".", "", "", [1, 0, 0]),
    "y": (".", "", "", [0, -1, 0]),
    "z": (".", "", "", [0, 0, -1]),
}

test_flatfield = np.array([[0, 0, 0], [0, 0, 0]])


@pytest.fixture
def dummy_singla_master_file():
    test_hdf_file = tempfile.NamedTemporaryFile(suffix=".h5", delete=True)
    with h5py.File(test_hdf_file, "w") as test_master_file:
        test_master_file["/entry/instrument/detector/count_time"] = 0.1
        test_master_file[
            "/entry/instrument/detector/detectorSpecific/data_collection_date"
        ] = "2023-12-06T10:30:42.039+02:00"
        test_master_file["/entry/instrument/detector/description/"] = b"Dectris Singla"
        test_master_file["/entry/instrument/detector/pixel_mask_applied"] = False
        test_master_file[
            "/entry/instrument/detector/detectorSpecific/flatfield"
        ] = test_flatfield
        test_master_file[
            "/entry/instrument/detector/detectorSpecific/software_version"
        ] = b"0.0.0"
        test_master_file["/entry/instrument/detector/nimages"] = 100
    yield test_hdf_file


def test_isSingla_master_file(dummy_singla_master_file):
    assert SinglaMaster.isDectrisSingla(dummy_singla_master_file.name)


def test_SignalMaster_get_mask_and_flatfiled(dummy_singla_master_file):
    with h5py.File(dummy_singla_master_file.name, "r") as fh:
        master = SinglaMaster(fh)
        mask_info = master.get_mask()
        flatfield_info = master.get_flatfield()

    assert not mask_info[0]  # mask applied is false
    assert mask_info[1] is None
    assert not flatfield_info[0]  # Flatfield applied is false
    assert_array_equal(flatfield_info[1], test_flatfield)


def test_get_mask_and_flatfield_from_singla_master_file(dummy_singla_master_file):
    D = extract_detector_info_from_master(dummy_singla_master_file.name)
    assert "pixel_mask" in D.keys() and "flatfield" in D.keys()
    assert D["pixel_mask"] is None
    assert D["pixel_mask_applied"] == 0
    assert np.all(D["flatfield"] == test_flatfield)
    assert D["flatfield_applied"] is False


def test_get_software_version_from_singla_master_file(dummy_singla_master_file):
    D = extract_detector_info_from_master(dummy_singla_master_file.name)
    assert "software_version" in D.keys()
    assert isinstance(D["software_version"], bytes)
    assert D["software_version"].decode() == "0.0.0"


def test_get_exposure_time_from_singla_master_file(dummy_singla_master_file):
    exp_time = extract_exposure_time_from_master(dummy_singla_master_file.name)
    assert isinstance(exp_time, float)
    assert exp_time == 0.1


def test_get_collection_start_time_from_singla_master_file(dummy_singla_master_file):
    start_time = extract_start_time_from_master(dummy_singla_master_file.name)
    assert isinstance(start_time, datetime)
    assert start_time == datetime.fromisoformat("2023-12-06T10:30:42")


def test_get_nimages_and_triggers(dummy_singla_master_file):
    with h5py.File(dummy_singla_master_file.name, "r") as fh:
        master = SinglaMaster(fh)
        nimages = master.get_number_of_images()
        ntriggers = master.get_number_of_triggers()
        mode = master.get_trigger_mode()
    assert nimages is not None
    assert ntriggers is None and mode is None


def test_centroid_max_calculation():
    test_image = np.array([[0, 1, 2], [3, 4, 5]])
    x, y = centroid_max(test_image)
    assert (x, y) == (2.0, 1.0)


def test_find_beam_center_if_no_pixel_mask(dummy_singla_master_file):
    beam_center = find_beam_centre(dummy_singla_master_file.name, "")
    assert beam_center is None


def test_find_beam_center(dummy_singla_master_file):
    # Image and gap size of a Singla detector are the same as a Eiger 1M
    # For the purposes of this test, call build_an_eiger
    image_size = (1062, 1028)
    test_mask = build_an_eiger(image_size, "eiger 1M")
    idx0, idx1 = np.where(test_mask != 0)
    for i, j in zip(idx0, idx1):
        test_mask[i][j] = 1
    with h5py.File(dummy_singla_master_file.name, "w") as fh:
        fh["/entry/instrument/detector/detectorSpecific/pixel_mask"] = test_mask
    test_img = np.zeros((2, *image_size), dtype=np.uint16)
    test_img_hdf_file = tempfile.NamedTemporaryFile(suffix=".h5")
    with h5py.File(test_img_hdf_file, "w") as img:
        # singla specs are the same as eiger 1m so use that for simulation
        img["entry/data/data"] = test_img
    beam_center = find_beam_centre(
        dummy_singla_master_file.name, test_img_hdf_file.name
    )
    assert isinstance(beam_center, tuple)
    assert beam_center[0] is not None
    assert beam_center[1] is not None



import argparse

import pytest

from nexgen import __version__
from nexgen.command_line import version_parser


def test_version():
    assert __version__


def test_version_parser(capsys):
    for flag in "-V", "--version":
        with pytest.raises(SystemExit, match="0"):
            version_parser.parse_args([flag])
        assert f"NeXus generation tools {__version__}" in capsys.readouterr().out


def test_version_parser_is_optional():
    assert version_parser.parse_args([]) == argparse.Namespace()


def test_version_parser_has_no_help_flag(capsys):
    with pytest.raises(SystemExit, match="2"):
        version_parser.parse_args(["-h"])
    assert "error: unrecognized arguments: -h" in capsys.readouterr().err



"""
Sample definition utilities.
"""
from __future__ import annotations

from dataclasses import dataclass

from dataclasses_json import DataClassJsonMixin


@dataclass
class Sample(DataClassJsonMixin):
    name: str | None = None
    depends_on: str | None = None
    temperature: str | None = None



from numpy.testing import assert_array_equal

from nexgen.nxs_utils.Axes import Axis, TransformationType
from nexgen.utils import Point3D

test_rot_axis = Axis("omega", ".", TransformationType.ROTATION, Point3D(0, 0, -1), -90)
test_transl_axis = Axis(
    "sam_x", "omega", TransformationType.TRANSLATION, (1, 0, 0), 0, 0.1, 10
)


def test_axes():
    assert test_rot_axis.name == "omega"
    assert test_rot_axis.start_pos == -90

    assert test_transl_axis.name == "sam_x"
    assert test_transl_axis.depends == "omega"
    assert_array_equal(test_transl_axis.vector, (1, 0, 0))


def test_axis_units():
    assert test_rot_axis.transformation_type == "rotation"
    assert test_rot_axis.units == "deg"

    assert test_transl_axis.transformation_type == "translation"
    assert test_transl_axis.units == "mm"


def test_axis_with_Point3D_vector_input():
    assert type(test_rot_axis.vector) is tuple
    assert_array_equal(test_rot_axis.vector, (0, 0, -1))


def test_axis_is_scan():
    assert test_rot_axis.is_scan is False
    assert test_transl_axis.is_scan is True


def test_axis_start_pos():
    assert test_rot_axis.start_pos != 0.0
    assert test_rot_axis.start_pos == -90.0
    test_rot_axis_2 = Axis("phi", ".", TransformationType.ROTATION, Point3D(0, 0, -1))
    assert test_rot_axis_2.start_pos == 0.0


def test_axis_end_pos():
    assert test_rot_axis.end_pos == test_rot_axis.start_pos
    assert test_transl_axis.end_pos == 0.9



"""
Writer for NeXus format files.
"""
from __future__ import annotations

import logging
from pathlib import Path
from typing import Any, Dict, List, Tuple

import numpy as np
from freephil.common import scope_extract as ScopeExtract

from ..nxs_utils import (
    Attenuator,
    Axis,
    Beam,
    Detector,
    DetectorType,
    EigerDetector,
    Facility,
    Goniometer,
    Source,
    TransformationType,
    TristanDetector,
)
from ..nxs_utils.Detector import EIGER_CONST, UnknownDetectorTypeError
from ..nxs_write.NXmxWriter import EventNXmxFileWriter, NXmxFileWriter
from ..utils import coerce_to_path, coord2mcstas, imgcif2mcstas

logger = logging.getLogger("nexgen.cli_utils")


def add_tristan_spec(detector: ScopeExtract, tristanSpec: ScopeExtract):
    """
    Add metadata specific to LATRD Tristan to detector scope.

    Args:
        detector (scope_extract):      Scope defining the detector
        tristanSpec (scope_extract):   Scope defining Tristan specific input
    """
    for k, v in tristanSpec.__dict__.items():
        if "__phil" in k:
            continue
        detector.__inject__(k, v)


def phil2dict(D: Dict):
    l: List = [k for k in D.keys() if "__phil_" in k]
    for i in l:
        D.__delitem__(i)
    return D


def split_arrays(axes_names: List, array: List) -> Dict[str, Tuple]:
    """Split a list of values into arrays.

    This function splits up the list of values passed as input (eg. phil parameters, dictionary) \
    for vector, offset for all existing axes.

    Args:
        axes_names (List): Axes names.
        array (List): Array of values to be split up. It must be

    Raises:
        ValueError: When each axes doesn't have a corresponding array of size 3.

    Returns:
        array_dict (Dict[str, Tuple]): Dictionary of arrays corresponding to each axis. Keys are axes names.
    """
    array_dict = {}
    if len(axes_names) == len(array):
        array_dict = {ax: tuple(v) for ax, v in zip(axes_names, array)}
        return array_dict
    elif len(array) == 3 * len(axes_names):
        for j in range(len(axes_names)):
            a = array[3 * j : 3 * j + 3]
            array_dict[axes_names[j]] = tuple(a)
        return array_dict
    else:
        error_msg = (
            f"Number of axes {len(axes_names)} doesn't match the lenght of the array list {len(array)}."
            "Please check again and make sure that all axes have a matching array of size 3."
        )
        logger.error(error_msg)
        raise ValueError(error_msg)


def reframe_arrays(
    goniometer: Dict[str, Any],
    detector: Dict[str, Any],
    module: Dict[str, Any],
    coordinate_frame: str = "mcstas",
    new_coord_system: Dict[str, Any] = None,
):
    """
    Split a list of offset/vector values into arrays. If the coordinate frame is not mcstas, \
    convert the arrays using the base vectors of the new coordinate system.

    Args:
        goniometer (Dict[str, Any]): Goniometer geometry description.
        detector (Dict[str, Any]): Detector specific parameters and its axes.
        module (Dict[str, Any]): Geometry and description of detector module.
        coordinate_frame (str, optional): Coordinate system being used. If "imgcif", there's no need to pass a \
            new coordinate system definition, as the conversion is already included in nexgen. Defaults to "mcstas".
        new_coord_system (Dict[str, Any], optional): Definition of the current coordinate system. \
            It should at least contain a string defining the convention, origin and axes information as a tuple of (depends_on, type, units, vector). \
            e.g. for X axis: {"x": (".", "translation", "mm", [1,0,0])}. \
            Defaults to None.

    Raises:
        ValueError: When the input coordinate system name and the coordinate system convention for the vectors doesn't match.
    """
    # If the arrays of vectors/offsets are not yet split, start by doing that
    goniometer["vectors"] = list(
        split_arrays(goniometer["axes"], goniometer["vectors"]).values()
    )
    goniometer["offsets"] = list(
        split_arrays(goniometer["axes"], goniometer["offsets"]).values()
    )

    detector["vectors"] = list(
        split_arrays(detector["axes"], detector["vectors"]).values()
    )

    if "offsets" in module.keys():
        module["offsets"] = list(
            split_arrays(["fast_axis", "slow_axis"], module["offsets"]).values()
        )

    # Now proceed with conversion if needed
    if coordinate_frame.lower() != "mcstas":
        if coordinate_frame.lower() == "imgcif":
            # Goniometer
            goniometer["vectors"] = [imgcif2mcstas(v) for v in goniometer["vectors"]]
            goniometer["offsets"] = [imgcif2mcstas(v) for v in goniometer["offsets"]]

            # Detector
            detector["vectors"] = [imgcif2mcstas(v) for v in detector["vectors"]]

            # Module
            module["fast_axis"] = imgcif2mcstas(module["fast_axis"])
            module["slow_axis"] = imgcif2mcstas(module["slow_axis"])
            if "offsets" in module.keys():
                module["offsets"] = [imgcif2mcstas(off) for off in module["offsets"]]
        else:
            if coordinate_frame != new_coord_system["convention"]:
                error_msg = (
                    "The input coordinate frame value doesn't match the current cordinate system convention."
                    "Impossible to convert to mcstas."
                )
                logger.error(error_msg)
                raise ValueError(error_msg)
            mat = np.array(
                [
                    new_coord_system["x"][-1],
                    new_coord_system["y"][-1],
                    new_coord_system["z"][-1],
                ]
            )

            # Goniometer
            goniometer["vectors"] = [
                coord2mcstas(v, mat) for v in goniometer["vectors"]
            ]
            goniometer["offsets"] = [
                coord2mcstas(v, mat) for v in goniometer["offsets"]
            ]

            # Detector
            detector["vectors"] = [coord2mcstas(v, mat) for v in detector["vectors"]]

            # Module
            module["fast_axis"] = coord2mcstas(module["fast_axis"], mat)
            module["slow_axis"] = coord2mcstas(module["slow_axis"], mat)
            if "offsets" in module.keys():
                module["offsets"] = [
                    coord2mcstas(off, mat) for off in module["offsets"]
                ]


def _update_detector_constants(det_params: DetectorType, params: Dict):
    for k in list(det_params.constants.keys()):
        if k in list(params.keys()):
            det_params[k] = params[k]


def set_detector_params(
    det_name: str, use_meta: bool = False, **params
) -> DetectorType:
    """Set up the detector parameters from parsed input

    Args:
        det_name (str): Detector name.
        use_meta (bool, optional): Whether a meta file is present. Mostly important for Eiger links.

    Keyword Args:
        Required parameters to set up detector.
        For Eiger: {"description", "image_size", "sensor_material", "overload", "underload"}
        For Tristan: {"description", "image_size"}
        Plus any additional arguments fo overwrite constants.

    Raises:
        ValueError: When missing required tristan kwargs.
        ValueError: When missing required eiger kwargs.
        UnknownDetectorTypeError: Id a detector drifferent from tristan or eiger has been requested.

    Returns:
        DetectorType: Parameters which define the detector in use.
    """
    if "tristan" in det_name.lower():
        required_params = {"description", "image_size"}
        if not required_params.issubset(params.keys()):
            logger.error("Missing detector parameters information for Tristan.")
            raise ValueError("Missing tristan information.")
        det_params = TristanDetector(params["description"], params["image_size"])
        if len(params.keys()) > 2:
            _update_detector_constants(det_params, params)
    elif "eiger" in det_name.lower():
        required_params = {
            "description",
            "image_size",
            "sensor_material",
            "overload",
            "underload",
        }
        if not required_params.issubset(params.keys()):
            logger.error("Missing detector parameters information for Eiger.")
            raise ValueError("Missing eiger information.")
        det_params = EigerDetector(
            params["description"],
            params["image_size"],
            params["sensor_material"],
            params["overload"],
            params["underload"],
        )
        if use_meta is False and len(params.keys()) > 5:
            _update_detector_constants(det_params, params)
        if use_meta is False and len(params.keys()) == 5:
            for k in list(det_params.constants.keys()):
                det_params.constants[k] = None
    else:
        error_msg = "This CLI functionality is currently only available for Eiger and Tristan detectors."
        logger.error(error_msg)
        raise UnknownDetectorTypeError(error_msg)
    return det_params


# Write NeXus base classes
def call_writers(
    nxsfile: Path | str,
    datafiles: List[Path | str],
    coordinate_frame: str,
    data_type: Tuple[str, int],
    goniometer: Dict[str, Any],
    detector: Dict[str, Any],
    module: Dict[str, Any],
    source: Dict[str, Any],
    beam: Dict[str, Any],
    attenuator: Dict[str, Any],
    metafile: bool = False,
    timestamps: Tuple[str, str] = None,
    notes: Dict[str, Any] = None,
):
    """
    Call the writers for the NeXus base classes.

    Args:
        nxsfile (Path | str): NeXus file to be written.
        datafiles (List[Path | str]): List of at least 1 Path object to a HDF5 data file.
        coordinate_frame (str): Coordinate system being used. Accepted frames are imgcif and mcstas.
        data_type (Tuple[str, int]): Images or event-mode data, and eventually how many are being written.
        goniometer (Dict[str, Any] Goniometer geometry description.
        detector (Dict[str, Any]): Detector specific parameters and its axes.
        module (Dict[str, Any]): Geometry and description of detector module.
        source (Dict[str, Any]): Facility information.
        beam (Dict[str, Any]): Beam properties.
        attenuator (Dict[str, Any]): Attenuator properties.
        metafile (bool, optional): Whether a metafile is present. Defaults to False.
        timestamps (Tuple[str], optional): Start and end collection timestamps in ISO format. Defaults to None.
        notes (Dict, optional): Any additional information to write as NXnote. Defaults to None.
    """
    logger.info("Calling the writer ...")
    nxsfile = coerce_to_path(nxsfile)

    # Split vectors and offsets in goniometer and detector for writing
    reframe_arrays(
        goniometer,
        detector,
        module,
        coordinate_frame,
    )

    # Check that filenames are paths
    datafiles = [coerce_to_path(f) for f in datafiles]

    # Set up Source, Beam, Attenuator
    attenuator_new = Attenuator(attenuator["transmission"])
    beam_new = Beam(beam["wavelength"], beam["flux"])
    facility = Facility(
        source["name"], source["short_name"], source["type"], source["facility_id"]
    )
    source_new = Source(source["beamline_name"], facility, source["probe"])

    # Other params that might have been passed
    if "eiger" in detector["description"].lower():
        contants = {
            k: detector[k]
            for k in list(EIGER_CONST.keys())
            if k in list(detector.keys())
        }
    else:
        contants = {}

    # Set up detector
    detector_params = set_detector_params(
        detector["description"],
        metafile,
        image_size=detector["image_size"],
        sensor_material=detector["sensor_material"],
        overload=detector["overload"],
        underload=["underload"],
        **contants,
    )
    det_axes = []
    for idx, ax in enumerate(detector["axes"]):
        transf = (
            TransformationType.TRANSLATION
            if "translation" == detector["types"][idx]
            else TransformationType.ROTATION
        )
        a = Axis(
            ax,
            detector["depends"][idx],
            transf,
            detector["vectors"][idx],
            start_pos=detector["starts"][idx],
        )
        det_axes.append(a)
    det = Detector(
        detector_params,
        det_axes,
        detector["beam_center"],
        detector["exposure_time"],
        [module["fast_axis"], module["slow_axis"]],
    )

    # Set up gonio
    gonio_axes = []
    for idx, ax in enumerate(goniometer["axes"]):
        transf = (
            TransformationType.TRANSLATION
            if "translation" == goniometer["types"][idx]
            else TransformationType.ROTATION
        )
        num = 0 if goniometer["increments"][idx] == 0 else data_type[1]
        a = Axis(
            ax,
            goniometer["depends"][idx],
            transf,
            goniometer["vectors"][idx],
            start_pos=goniometer["starts"][idx],
            increment=goniometer["increments"][idx],
            num_steps=num if data_type[0] == "images" else 0,
        )
        gonio_axes.append(a)

    # Scan can be identified by goniometer functionality
    gonio = Goniometer(gonio_axes)

    # Aaaand write
    if data_type[0] == "images":
        writer = NXmxFileWriter(
            nxsfile,
            gonio,
            det,
            source_new,
            beam_new,
            attenuator_new,
            data_type[1],
        )
        writer.write(image_datafiles=datafiles, start_time=timestamps[0])
        writer.write_vds()
    else:
        writer = EventNXmxFileWriter(
            nxsfile, gonio, det, source_new, beam_new, attenuator_new
        )
        writer.write(start_time=timestamps[0])

    if notes:
        writer.add_NXnote(notes, "/entry/notes")

    if timestamps[1]:
        writer.update_timestamps(timestamps[1])



import tempfile
from unittest.mock import MagicMock

import h5py
import numpy as np
import pytest

from nexgen.nxs_utils import (
    Attenuator,
    Axis,
    Beam,
    Detector,
    EigerDetector,
    Goniometer,
    Source,
    TransformationType,
)
from nexgen.nxs_write.NXmxWriter import NXmxFileWriter

dummy_config = """{
    "nimages": 90,
    "ntrigger": 1,
    "omega_increment": 1.0,
    "omega_start": 0.0,
    "phi_increment": 0.0,
    "phi_start": 0.0,
    "two_theta_start": 10.,
    "two_theta_increment": 0.0,
    }"""


@pytest.fixture
def mock_goniometer() -> Goniometer:
    return Goniometer(
        [
            Axis(
                "omega",
                ".",
                TransformationType.ROTATION,
                (-1, 0, 0),
                start_pos=0,
                increment=1,
                num_steps=90,
            ),
            Axis("sam_z", "omega", TransformationType.TRANSLATION, (0, -1, 0)),
            Axis("sam_y", "sam_z", TransformationType.TRANSLATION, (-1, 0, 0)),
        ],
        {"omega": np.arange(0, 90, 1)},
    )


@pytest.fixture
def mock_source() -> Source:
    return Source("I03")


@pytest.fixture
def mock_beam() -> Beam:
    return Beam(wavelength=0.6)


@pytest.fixture
def mock_attenuator() -> Attenuator:
    return Attenuator(transmission=10.0)


@pytest.fixture
def mock_eiger() -> Detector:
    eiger_params = EigerDetector(
        "Eiger 2X 9M",
        [3262, 3108],
        "CdTe",
        50649,
        -1,
    )
    det_axes = [
        Axis("det_z", ".", TransformationType.TRANSLATION, (0, 0, 1), start_pos=500),
    ]
    return Detector(
        eiger_params,
        det_axes,
        [1590.7, 1643.7],
        0.01,
        [(1, 0, 0), (0, 1, 0)],
    )


@pytest.fixture
def dummy_eiger_meta_file():
    test_hdf_file = tempfile.TemporaryFile()
    test_meta_file = h5py.File(test_hdf_file, "w")
    test_meta_file["config"] = dummy_config
    test_meta_file["_dectris/nimages"] = np.array([90])
    test_meta_file["_dectris/ntrigger"] = np.array([1])
    test_meta_file["_dectris/wavelength"] = np.array([0.6])
    test_meta_file["_dectris/x_pixels_in_detector"] = np.array([3180])
    test_meta_file["_dectris/y_pixels_in_detector"] = np.array([3262])
    test_meta_file["_dectris/detector_distance"] = np.array([0.19])
    test_meta_file["_dectris/bit_depth_readout"] = np.array([32])
    test_meta_file["flatfield"] = np.array([[0, 0, 0]])
    test_meta_file["_dectris/software_version"] = np.string_("0.0.0")
    test_meta_file["mask"] = np.array([[0, 1, 1], [1, 0, 0]])
    test_meta_file["_dectris/pixel_mask_applied"] = False
    yield test_meta_file


@pytest.fixture
def dummy_nexus_file():
    test_hdf_file = tempfile.TemporaryFile()
    test_nexus_file = h5py.File(test_hdf_file, "w")
    yield test_nexus_file


@pytest.fixture
def dummy_NXmxWriter():
    with tempfile.NamedTemporaryFile(suffix=".nxs", delete=True) as test_nxs_file:
        test_fw = NXmxFileWriter(
            test_nxs_file.name,
            MagicMock(),
            MagicMock(),
            MagicMock(),
            MagicMock(),
            MagicMock(),
            10,
        )
        yield test_fw



"""
Create a NeXus file for serial crystallography datasets collected on Tristan10M detector either on I19-2 or I24 beamlines.
"""
from __future__ import annotations

import logging
from collections import namedtuple
from pathlib import Path

from .. import log
from ..nxs_utils import Attenuator, Beam, Detector, Goniometer, Source, TristanDetector
from ..nxs_write.NXmxWriter import EventNXmxFileWriter
from ..utils import Point3D, find_in_dict, get_iso_timestamp
from .beamline_utils import collection_summary_log

# Define a logger object and a formatter
logger = logging.getLogger("nexgen.SSX_Tristan")

ssx_tr_collect = namedtuple(
    "ssx_collect",
    [
        "exposure_time",
        "detector_distance",
        "beam_center",
        "transmission",
        "wavelength",
        "start_time",
        "stop_time",
        "chipmap",
        "chip_info",
    ],
)

ssx_tr_collect.__doc__ = (
    """Parameters that define a serial collection using a Tristan detector."""
)


def ssx_tristan_writer(
    visitpath: Path | str,
    filename: str,
    beamline: str,
    **ssx_params,
):
    """
    Gather all parameters from the beamline and call the NeXus writers.

    Args:
        visitpath (Path | str): Path to colection directory.
        filename (str): Root of the filename.
        beamline (str): Beamline on which the experiment is being run.

    Keyword Args:
        exp_time (float): Exposure time, in s.
        det_dist (float): Distance between sample and detector, in mm.
        beam_center (List[float, float]): Beam center position, in pixels.
        transmission (float): Attenuator transmission, in %.
        wavelength (float): Wavelength of incident beam, in A.
        flux (float): Total flux.
        start_time (datetime): Experiment start time.
        stop_time (datetime): Experiment end time.
        chip_info (Dict): For a grid scan, dictionary containing basic chip information.
            At least it should contain: x/y_start, x/y number of blocks and block size, x/y number of steps and number of exposures.
        chipmap (Path | str): Path to the chipmap file corresponding to the experiment,
            or 'fullchip' indicating that the whole chip is being scanned.
    """
    if not find_in_dict("start_time", ssx_params):
        ssx_params["start_time"] = None
    if not find_in_dict("stop_time", ssx_params):
        ssx_params["stop_time"] = None

    # Get info from the beamline
    SSX_TR = ssx_tr_collect(
        exposure_time=float(ssx_params["exp_time"])
        if find_in_dict("exp_time", ssx_params)
        else None,
        detector_distance=float(ssx_params["det_dist"])
        if find_in_dict("det_dist", ssx_params)
        else None,
        beam_center=ssx_params["beam_center"]
        if find_in_dict("beam_center", ssx_params)
        else (0, 0),
        transmission=float(ssx_params["transmission"])
        if find_in_dict("transmission", ssx_params)
        else None,
        wavelength=float(ssx_params["wavelength"])
        if find_in_dict("wavelength", ssx_params)
        else None,
        start_time=ssx_params["start_time"].strftime("%Y-%m-%dT%H:%M:%S")
        if ssx_params["start_time"]
        else None,  # This should be datetiem type
        stop_time=ssx_params["stop_time"].strftime("%Y-%m-%dT%H:%M:%S")
        if ssx_params["stop_time"]
        else None,  # idem.
        chipmap=ssx_params["chipmap"] if find_in_dict("chipmap", ssx_params) else None,
        chip_info=ssx_params["chip_info"]
        if find_in_dict("chip_info", ssx_params)
        else None,
    )

    visitpath = Path(visitpath).expanduser().resolve()
    filename = ssx_params["filename"]

    logfile = SSX_TR.visitpath / f"{beamline}_TristanSSX_nxs_writer.log"
    # Configure logging
    log.config(logfile.as_posix())

    logger.info(
        f"Start NeXus File Writer for time-resolved SSX on beamline {beamline} at DLS."
    )

    logger.info(f"Current collection directory: {visitpath.as_posix()}")
    # Get NeXus filename
    master_file = visitpath / f"{filename}.nxs"
    logger.info("NeXus file will be saved as %s" % master_file.as_posix())

    # Check if metafile in directory.
    try:
        metafile = [
            f for f in visitpath.iterdir() if filename + "_meta" in f.as_posix()
        ][0]
        logger.info(f"Found {metafile} in directory.")
    except IndexError:
        logger.warning(
            "Missing metadata file at this time, something might be wrong with this collection."
            f"The hard link to {filename}_meta.h5 will most likely be broken."
        )

    # Get parameters depending on beamline
    logger.info(f"DLS Beamline: {beamline.upper()}.")
    if "I19" in beamline.upper():
        source = Source("I19-2")
        from .I19_2_params import I19_2Tristan as axes_params

    elif "I24" in beamline.upper():
        source = Source("I19-2")
        from .I24_params import I24Eiger as axes_params

        axes_params.fast_axis = Point3D(-1, 0, 0)
        axes_params.slow_axis = Point3D(0, 1, 0)

    else:
        raise ValueError(
            "Unknown beamline for SSX collections with Tristan 10M detector."
            "Beamlines currently enabled for the writer: I24, I19-2."
        )

    # Define Attenuator
    attenuator = Attenuator(SSX_TR.transmission)
    # Define Beam
    wl = SSX_TR.wavelength
    flux = ssx_params["flux"] if "flux" in ssx_params.keys() else None
    beam = Beam(wl, flux)

    # Define Detector axes
    det_axes = axes_params.det_axes

    # Define Detector
    tristan_params = TristanDetector("Tristan 10M", (3043, 4183))
    if "I19" in beamline.upper():
        det_axes[0].start_pos = 0.0  # two_theta
        det_axes[1].start_pos = SSX_TR.detector_distance  # det_z
    else:
        # Only det_z for detector axes
        det_axes[0].start_pos = SSX_TR.detector_distance

    detector = Detector(
        tristan_params,
        det_axes,
        SSX_TR.beam_center,
        SSX_TR.exposure_time,
        [axes_params.fast_axis, axes_params.slow_axis],
    )

    # Define Goniometer axes
    gonio_axes = axes_params.gonio

    # Get rotation scan range array and axis
    osc_axis = "phi" if "I19" in SSX_TR.location else "omega"
    osc_range = (0.0, 0.0)

    OSC = {osc_axis: osc_range}

    # Define Goniometer
    goniometer = Goniometer(gonio_axes, OSC)

    # Get timestamps in the correct format
    timestamps = (
        get_iso_timestamp(SSX_TR.start_time),
        get_iso_timestamp(SSX_TR.stop_time),
    )

    collection_summary_log(
        logger,
        goniometer,
        detector,
        attenuator,
        beam,
        source,
        timestamps,
    )

    try:
        EventFileWriter = EventNXmxFileWriter(
            master_file,
            goniometer,
            detector,
            source,
            beam,
            attenuator,
        )
        EventFileWriter.write(start_time=timestamps[0])
        if timestamps[1]:
            # TODO add calc for estimated and put it here (same for SSX_eiger)
            EventFileWriter.update_timestamps(timestamps[1], "end_time")

        # Save chipmap (list of city blocks)
        if SSX_TR.chipmap:
            # Assuming 8x8 fast chip
            from .SSX_chip import read_chip_map

            chip = read_chip_map(SSX_TR.chipmap, 8, 8)
            mapping = {"chipmap": str(chip)}
            logger.info(
                f"Chipmap read from {SSX_TR.chipmap}, saving in '/entry/source/notes/chipmap'."
            )
            EventFileWriter.add_NXnote(mapping, "/entry/source/notes/")
            # To read this: eval(dset[()])

        # Save chip info if passed. If not save I24-like chip info plus warning message
        logger.info("Save chip information in /entry/source/notes/chip")
        if SSX_TR.chip_info:
            # Make chip info more readable
            chip_info = {k: v[1] for k, v in SSX_TR.chip_info.items()}
            chipdef = {"chip": str(chip_info)}
            EventFileWriter.add_NXnote(chipdef, "/entry/source/notes/")
        else:
            logger.warning(
                "Dictionary containing chip info was not passed to the writer."
                "The following values will be written as default: "
                "x/y_num_blocks = 8 \n x/y_block_size = 3.175 \n x/y_num_steps = 20 \n x/y_step_size = 0.125"
            )
            from .SSX_chip import CHIP_DICT_DEFAULT as chip_info

            chipdef = {"chip": str(chip_info)}
            EventFileWriter.add_NXnote(chipdef, "/entry/source/notes/")

            logger.info(f"The file {master_file} was written correctly.")
    except Exception as err:
        logger.exception(err)
        logger.info(
            f"An error occurred and {master_file} couldn't be written correctly."
        )
        raise



"""
IO tool to gather beamline and collection information from xml file.
"""
from __future__ import annotations

import xml.etree.ElementTree as ET
from pathlib import Path
from typing import Dict, List, Tuple


class ExtendedRequestIO:
    """Define an ExtendedRequest object which in GDA gathers all the information regarding beamline and collection into an xml file."""

    def __init__(self, xmlfile: Path | str):
        self.tree = ET.parse(xmlfile)
        self.root = self.tree.getroot()

    def getCollectionInfo(self) -> Tuple[str, str, str]:
        directory = self.root.find(".//directory").text
        prefix = self.root.find(".//prefix").text
        run_number = self.root.find(".//runNumber").text
        return directory, prefix, run_number

    def getOscillationSequence(self) -> Dict[str, float]:
        osc_seq_node = self.root.find(".//oscillation_sequence")
        osc_seq = {
            "start": float(osc_seq_node.find(".//start").text),
            "range": float(osc_seq_node.find(".//range").text),
            "number_of_images": float(osc_seq_node.find(".//number_of_images").text),
            "exposure_time": float(osc_seq_node.find(".//exposure_time").text),
            "number_of_passes": float(osc_seq_node.find(".//number_of_passes").text),
        }
        return osc_seq

    def getAxisChoice(self) -> str:
        ax = self.root.find(".//axisChoice").text
        return ax.lower()

    def getOtherAxis(self) -> float:
        val = float(self.root.find(".//otherAxis").text)
        return val

    def getKappa(self) -> float:
        return float(self.root.find(".//kappa").text)

    def getChi(self) -> float:
        return float(self.root.find(".//chi").text)

    def getTwoTheta(self) -> float:
        return float(self.root.find(".//twoTheta").text)

    def getSampleDetectorDistance(self) -> float:
        return float(self.root.find(".//sampleDetectorDistanceInMM").text)

    def getTransmission(self) -> float:
        return float(self.root.find(".//transmissionInPerCent").text)

    def getResolution(self) -> float:
        return float(self.root.find(".//resolution").text)


def read_scan_from_xml(ecr: ExtendedRequestIO):
    """
    Extract information about the scan contained in the xml file.

    Args:
        ecr (ExtendedRequestIO): XML tree parser.
        xmlfile (Path | str): Path to xml file.

    Returns:
        scan_axis (str): Name of the rotation scan axis
        pos (Dict): Dictionary containing the (start,end,increment) values for each goniometer axis.
        num (int): Number of images written.
    """
    # Goniometer
    osc_seq = ecr.getOscillationSequence()
    # Find scan range
    if osc_seq["range"] == 0.0:
        scan_range = (osc_seq["start"], osc_seq["start"])
        num = osc_seq["number_of_images"]
    else:
        start = osc_seq["start"]
        num = osc_seq["number_of_images"]
        stop = start + num * osc_seq["range"]
        scan_range = (start, stop)
    # Determine scan axis
    if ecr.getAxisChoice() == "omega":
        scan_axis = "omega"
        omega_pos = (*scan_range, osc_seq["range"])  # 0.0)
        phi_pos = (*2 * (ecr.getOtherAxis(),), 0.0)
    else:
        scan_axis = "phi"
        phi_pos = (*scan_range, osc_seq["range"])  # 0.0)
        omega_pos = (*2 * (ecr.getOtherAxis(),), 0.0)
    kappa_pos = (*2 * (ecr.getKappa(),), 0.0)

    pos = {
        "omega": omega_pos,
        "phi": phi_pos,
        "kappa": kappa_pos,
        "sam_x": (0.0, 0.0, 0.0),
        "sam_y": (0.0, 0.0, 0.0),
        "sam_z": (0.0, 0.0, 0.0),
    }

    return scan_axis, pos, int(num)


def read_det_position_from_xml(
    ecr: ExtendedRequestIO,
    det_description: str,
) -> List[float]:
    """Extract the detector position contained in the xml file.

    Args:
        ecr (ExtendedRequestIO): XML tree parser.
        det_description (str): Detector description

    Returns:
        List[float]: Detector axes positions in the order [2theta, det_z]
    """
    if "tristan" in det_description.lower():
        return [0.0, ecr.getSampleDetectorDistance()]
    elif "eiger" in det_description.lower():
        return [ecr.getTwoTheta(), ecr.getSampleDetectorDistance()]
    return [0.0, ecr.getSampleDetectorDistance()]
